from pie_extended.models.fr.imports import get_iterator_and_processor
from pie_extended.testing_utils import FakeTagger
from typing import List, Tuple

from unittest import TestCase
from .test_lasla import make_fake_data


def make_controller(sentences: List[str]):
    # Add the lemmatizer routes
    tagger = FakeTagger(
        make_fake_data(sentences),
        tasks="lemma,MODE,TEMPS,PERS,NOMB,GENRE,CAS,DEGRE,POS".split(",")
    )
    iterator, processor = get_iterator_and_processor()
    return tagger, iterator, processor


class TestFr(TestCase):
    """
    def test_elision_apostrophe(self):
        string = "q'il meurt"
        treated = ["q' il meurt"]
        tagger, it, pro = make_controller(treated)
        out = tagger.tag_str(string, it, pro)
        self.assertEqual(out[0]["form"], "q'")
        self.assertEqual(out[0]["treated"], "q")

    def test_elision_apostrophe_and_quote(self):
        string = "'q'il meurt 'dit il'"
        treated = ["' q' il meurt  ' dit il '"]
        tagger, it, pro = make_controller(treated)
        out = tagger.tag_str(string, it, pro)
        self.assertEqual(out[0]["form"], "'")
        self.assertEqual(out[0]["treated"], "'")
        self.assertEqual(out[1]["form"], "q'")
        self.assertEqual(out[1]["treated"], "q")
        self.assertEqual(out[-1]["form"], "'", "Last apostrophe is kept")
        # Ending and starting apostrophe are not reinserted for some reason.
    """

    def test_tokenization_clitics(self):
        iterator, _ = get_iterator_and_processor()
        iterator.tokenizer.replacer = lambda x: x  # Until model is fixed
        self.assertEqual(
            list(
                iterator.tokenizer.sentence_tokenizer(
                    "L'a-t-il mangé celui-là-même ou l'a-t-on rangé celui-ci ? -très mauvaise tokenization"
                )
            ),
            [
                ["L'", 'a', '-t-il', 'mangé', 'celui', '-là', '-même',
                 'ou', "l'", 'a', '-t-on', 'rangé', 'celui', '-ci', '?'],
                ["-", "très", "mauvaise", "tokenization"]
            ],
            "Dots around roman number are not sentences markers"
        )

    def test_tokenization_peut_etre(self):
        iterator, _ = get_iterator_and_processor()
        iterator.tokenizer.replacer = lambda x: x  # Until model is fixed
        self.assertEqual(
            list(
                iterator.tokenizer.sentence_tokenizer(
                    "Peut-être a-t-il raison ?"
                )
            ),
            [['Peut', '-', 'être', 'a', '-t-il', 'raison', '?']],
            "Dots around roman number are not sentences markers"
        )

    def test_tokenization_aujourdhui(self):
        iterator, _ = get_iterator_and_processor()
        iterator.tokenizer.replacer = lambda x: x  # Until model is fixed
        self.assertEqual(
            list(
                iterator.tokenizer.sentence_tokenizer(
                    "Aujourd'hui a-t-il raison ?"
                )
            ),
            [['Aujourd\'hui', 'a', '-t-il', 'raison', '?']],
            "Dots around roman number are not sentences markers"
        )

    def test_tokenization_elise_t_euphonique(self):
        """ Check that -t' are correctly tokenized """
        iterator, _ = get_iterator_and_processor()
        self.assertEqual(
            list(
                iterator.tokenizer.sentence_tokenizer("Va-t'en va-nu-pieds !")
            ),
            [
                ["Va", "-t'", "en", "va", "-", "nu", "-", "pieds", "!"]
            ]
        )

    def test_tokenization_abbreviations(self):
        """ Check that abbreviation are recognized """
        iterator, _ = get_iterator_and_processor()
        self.assertEqual(
            list(
                iterator.tokenizer.sentence_tokenizer("La Zoo. montre des limites. cf. un truc. Et V. n.  est un neutre"
                                                      ". Mais j'aime ma Lit. et mon lit.!")
            ),
            [
                ['La', 'Zoo.', 'montre', 'des', 'limites', '.'],
                ['cf.', 'un', 'truc', '.'],
                ['Et', 'V.', 'n.', 'est', 'un', 'neutre', '.'],
                ['Mais', "j'", 'aime', 'ma', 'Lit.', 'et', 'mon', 'lit', '.', '!']
            ]
        )



from pie_extended.pipeline.postprocessor.proto import ProcessorPrototype, ChainedProcessor
from typing import Optional, Dict, List
if "typing" == "nottyping":
    from ..tokenizers.memorizing import MemorizingTokenizer


class RuleBasedProcessor(ChainedProcessor):
    """ Applies rules found in rules(token_annotation)

    """

    def __init__(self, apply_on_reinsert: bool = False, head_processor: Optional[ProcessorPrototype] = None, **kwargs):
        """ Apply rules on output of the taggers

        :param apply_on_reinsert: Apply rules on reinsert task
        :param head_processor: Processor to use before post-processing its results

        >>> class ExampleRule(RuleBasedProcessor):
        ...     def rules(self, annotation: Dict[str, str]) -> Dict[str, str]:
        ...         if annotation["form"] == "need":
        ...             annotation["1"] = "REPLACED"
        ...         return annotation
        >>> processor = ExampleRule()
        >>> processor.set_tasks(["1", "2"])
        ['1', '2']
        >>> processor.get_dict("token", ["a", "b"]) == [{"form": "token", "1": "a", "2": "b"}]
        True
        >>> processor.get_dict("need", ["a", "b"]) == [{"form": "need", "1": "REPLACED", "2": "b"}]
        True
        """
        super(RuleBasedProcessor, self).__init__(head_processor=head_processor, **kwargs)
        self.apply_on_reinsert = apply_on_reinsert

    def rules(self, annotation: Dict[str, str]) -> Dict[str, str]:
        return annotation

    def reinsert(self, form: str) -> Dict[str, str]:
        anno = super(RuleBasedProcessor, self).reinsert(form)
        if self.apply_on_reinsert:
            return self.rules(anno)
        return anno

    def get_dict(self, token: str, tags: List[str]) -> List[Dict[str, str]]:
        return [self.rules(anno) for anno in self.head_processor.get_dict(token, tags)]



import regex as re
from typing import Dict, List, Optional, Generator, Tuple

from pie_extended.pipeline.postprocessor.proto import ChainedProcessor, ProcessorPrototype
from pie_extended.pipeline.postprocessor.glue import GlueProcessor
from pie_extended.pipeline.postprocessor.rulebased import RuleBasedProcessor
from pie_extended.utils import roman_number


class LatinRulesProcessor(RuleBasedProcessor):
    """ Lasla data has no punctuation, we tag it automatically.

    "ne" token can be two different lemma, but I don't remember why I wrote this part. (ne/nec ?)

    >>> p = LatinRulesProcessor()
    >>> p.set_tasks(["lemma", "pos", "morph", "treated"])
    ['lemma', 'pos', 'morph', 'treated']
    >>> p.get_dict("uinipollens", ['界pollens', '', '', 'uinipollens']) == [
    ...    {'lemma': 'pollens', 'treated': 'uinipollens', 'morph': 'MORPH=empty', 'pos': 'ADJqua',
    ...     'form': '{uinipollens}', 'Dis': '_'}
    ... ]
    True
    >>> p.get_dict("similist", ['界sum', '', '', 'similist']) == [{'lemma': 'sum1', 'treated': 'similist',
    ... 'morph': 'Numb=Sing|Mood=Ind|Tense=Pres|Voice=Act|Person=3',
    ... 'pos': 'VER', 'form': '{similist}', 'Dis': '_'}]
    True

    """
    PONCTU = re.compile(r"^\W+$")
    GREEK = re.compile(r"^\p{Greek}+$")
    CLITICS_POS = {
        'audeo': 'VER',
        'consultum': 'NOM',
        'cum': 'CON',
        'ientaculum': 'NOM',
        'ille': 'PROdem',
        'ipse': 'PROdem',
        'is': 'PROdem',
        'iste': 'PROdem',
        'ne': 'ADV',
        'pollens': 'ADJqua',
        'que': 'CON',
        'scribo': 'VER',
        'sum': 'VER',
        'ue': 'CON',
        'unus': '',
        'uolo': ''
    }

    CLITICS_MORPH = {
        'sum': 'Numb=Sing|Mood=Ind|Tense=Pres|Voice=Act|Person=3'
    }
    CLITICS_DIS = {
        'cum': '3',
        'ne': '2',
        'sum': '1'
    }

    def rules(self, annotation: Dict[str, str]) -> Dict[str, str]:
        # If Else condition
        token = annotation["form"]

        if annotation["lemma"].startswith("界"):
            lem = annotation["lemma"][1:]
            return {
                "lemma": lem + self.CLITICS_DIS.get(lem, ""),
                "treated": annotation["treated"],
                "morph": self.CLITICS_MORPH.get(lem, "MORPH=empty"),
                "pos": self.CLITICS_POS.get(lem, "UNK"),
                "form": "{"+annotation["form"]+"}",
                "Dis": "_"
            }

        if self.PONCTU.match(token):
            return {"form": token, "lemma": token, "pos": "PUNC", "morph": "MORPH=empty",
                    "treated": annotation['treated'], "Dis": "_"}
        elif self.GREEK.match(token):
            return {"form": token, "lemma": "[Greek]", "pos": "FOR", "morph": "MORPH=empty",
                    "treated": annotation['treated'], "Dis": "_"}
        elif token.startswith("[IGN:"):
            return {"form": token, "lemma": "[IGNORED]", "pos": "OUT", "morph": "MORPH=empty",
                    "treated": annotation['treated'], "Dis": "_"}
        elif token.startswith("[REF:"):
            return {"form": token, "lemma": "[METADATA]", "pos": "OUT", "morph": "MORPH=empty",
                    "treated": annotation['treated'], "Dis": "_"}
        elif annotation["lemma"].isdigit() and annotation["treated"].isdigit() and not token.isnumeric():
            try:
                annotation["lemma"] = str(roman_number(token))
            except KeyError:
                annotation["lemma"] = "<UNK_NUMBER>"
                print("Weird behavior on this token", annotation)
        return annotation

    def __init__(self, *args, **kwargs):
        super(LatinRulesProcessor, self).__init__(*args, **kwargs)

class LatinGlueProcessor(GlueProcessor):
    OUTPUT_KEYS = ["form", "lemma", "pos", "morph", "Dis"]
    GLUE = {"morph": ["Case", "Numb", "Gend", "Deg", "Mood", "Tense", "Voice", "Person"]}
    WHEN_EMPTY = {"morph": "MORPH=empty"}
    EMPTY_TAG: Dict[str, str] = {"Case": "_", "Numb": "_", "Deg": "_", "Mood": "_", "Tense": "_", "Voice": "_",
                                 "Person": "_", "Gend": "_"}

    def __init__(self, *args, **kwargs):
        super(LatinGlueProcessor, self).__init__(*args, **kwargs)


class MoodTenseVoice(ChainedProcessor):
    def __init__(self, head_processor: Optional[ProcessorPrototype],
                 empty_value: str = "_", **kwargs):
        super(MoodTenseVoice, self).__init__(head_processor=head_processor)

        self._out_tasks = []
        self.empty_value = empty_value

    def set_tasks(self, tasks):
        self._tasks = self.head_processor.set_tasks(tasks)
        self._out_tasks = [
            subtask
            for task in self._tasks
            for subtask in task.split("_")
        ]
        return self.tasks

    def reinsert(self, form: str) -> Dict[str, str]:
        return dict(form=form, **{key: self.empty_value for key in self._out_tasks if key != "form"})

    def _yield_key(self, dic: Dict[str, str]) -> Generator[Tuple[str, str], None, None]:
        for key, value in dic.items():
            if "_" in key:
                keys, values = key.split("_"), value.split("|")
                for k, v in zip(keys, values+[self.empty_value]*(len(keys)-len(values))):
                    yield k, v
            else:
                yield key, value

    def get_dict(self, token: str, tags: List[str]) -> List[Dict[str, str]]:
        return [
            dict(self._yield_key(dic))
            for dic in self.head_processor.get_dict(token, tags)
        ]

    def reset(self):
        self.head_processor.reset()




import os
import sys
from typing import Tuple, Iterable, List, Union
from importlib import import_module

import requests

from .. import models
from ..utils import Metadata, PATH, get_path
from ..tagger import ExtensibleTagger
from ..utils import ObjectCreator
from pie.utils import model_spec


def get_model(model: str):
    """ Retrieve a module given a string

    :param model: Module Name
    :return: Module
    """
    return import_module("{}.{}".format(models.__name__, model))


def get_imports(module):
    return import_module("{}.{}".format(module.__name__, "imports"))


def _download(url, filename):
    with open(filename, 'wb') as f:
        response = requests.get(url, stream=True)
        total = response.headers.get('content-length')

        if total is None:
            f.write(response.content)
        else:
            downloaded = 0
            total = int(total)
            for data in response.iter_content(chunk_size=max(int(total/1000), 1024*1024)):
                downloaded += len(data)
                f.write(data)
                done = int(50*downloaded/total)
                sys.stdout.write('\r[{}{}]'.format('█' * done, '.' * (50-done)))
                sys.stdout.flush()
    sys.stdout.write('\n')


def download(module: str) -> Iterable[Union[str, int]]:
    """ Download dependencies for the given module

    :param module: Module for which to download models and static files in general
    """
    lemmatizer = get_model(module)
    os.makedirs(os.path.join(PATH, module), exist_ok=True)
    yield len(lemmatizer.DOWNLOADS)
    for file in lemmatizer.DOWNLOADS:

        new_path = get_path(module, file.name)
        _download(file.url, new_path)
        yield file.name


def get_list() -> Iterable[Tuple[str, Metadata]]:
    """ Retrieve a list of available modules
    """
    for module in models.modules:
        desc = getattr(get_model(module), "DESC", None)
        if desc:
            yield module, desc


def get_tagger(model: str, batch_size: int = 16, device="cpu", model_path=None) -> ExtensibleTagger:
    """ Retrieve the tagger

    :param model: Module of the tagger
    :param batch_size: Size of the batch
    :param device: Device to use (cuda/cpu)
    :param model_path: Path to the model if you want to override the package one
    :return: Tagger
    """
    module = get_model(model)

    disambiguator = getattr(get_imports(module), "Disambiguator", None)
    if isinstance(disambiguator, ObjectCreator):
        disambiguator = disambiguator.create()
    tagger = ExtensibleTagger(disambiguation=disambiguator, batch_size=batch_size, device=device)
    model_spec_string = model_path or getattr(module, "Models")
    for model, tasks in model_spec(model_spec_string):
        tagger.add_model(model, *tasks)
    return tagger


def tag_file(
        model: str, tagger: ExtensibleTagger,
        fpath: str,
        reset_exclude_patterns: bool = False,
        exclude_patterns: List[str] = None,
        no_tokenizer: bool = False,
        max_tokens: int = 256):
    """ Tag a file with a given model

    :param model: Module name of the model
    :param tagger: Tagger that should be used
    :param fpath: Path to the file to edit
    :param reset_exclude_patterns: Remove all pre-registered token exclusion regular expressions
    :param exclude_patterns: New exclude patterns to add to the data iterator (Does not require reset)
    :param no_tokenizer: Does not use usual tokenizers (new line = word separator, two new lines = sentence_
    """
    module = get_model(model)
    iterator, processor = getattr(get_imports(module), "get_iterator_and_processor")(max_tokens=max_tokens)
    # Remove first pattern
    if reset_exclude_patterns:
        iterator.reset_patterns()

    # Add new
    if exclude_patterns:
        for pattern in exclude_patterns:
            iterator.add_pattern(pattern)

    tagger.tag_file(fpath, iterator=iterator, processor=processor, no_tokenizer=no_tokenizer)
    return True


def get_addons(model: str):
    """ Runs the `addons` function from a module """
    module = get_model(model)
    addons = getattr(get_imports(module), "addons", lambda: print("No add-ons needed for " + model))
    return addons()



from typing import Generator, List
import regex as re
import string
from pie.tagger import regexsplitter, SECTION, FULLSTOP

WORD = r'([{}])'.format(string.punctuation)

RE_BYPASS_SENTENCE = re.compile(r"(?:\r?\n){2,}")
RE_BYPASS_WORD = re.compile(r"(?:\r?\n)")


class SimpleTokenizer(object):
    """ Tokenizer that memoryze what it tokenized.

    Mostly used to normalized input as input time and then reinserting normalized input

    """
    def __init__(self):
        self.section = regexsplitter(SECTION)
        self.fullstop = regexsplitter(FULLSTOP)
        self.word = regexsplitter(WORD)

    def sentence_tokenizer(self, text: str, lower: bool = False) -> Generator[List[str], None, None]:
        for line in self.section(text):
            for sentence in self.fullstop(line):
                yield self.word_tokenizer(sentence, lower=lower)

    def word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        sentence = [w for raw in text.split() for w in self.word(raw)]
        if lower:
            sentence = [w.lower() for w in sentence]
        return sentence

    def reset(self):
        """Can be used between documents for example """
        pass

    def bypass_tokenizer(self, data: str, lower: bool = False) -> Generator[List[str], None, None]:
        """ Function to enable pretokenized input while using replaces or the likes

        :param data: The string (one new line = one word, two new lines = two words)
        :param lower: Whether to lower the input


        >>> tokenizer = SimpleTokenizer()
        >>> list(tokenizer.bypass_tokenizer("one\\ntwo\\nthree\\n\\n.//.\\na\\nz"))
        [['one', 'two', 'three'], ['.//.', 'a', 'z']]
        """

        for sentence in RE_BYPASS_SENTENCE.split(data):
            if sentence:
                yield [word for word in RE_BYPASS_WORD.split(sentence) if word]


class LengthTokenizer(SimpleTokenizer):
    def __init__(self, max_len=35):
        super(LengthTokenizer, self).__init__()
        self.max_len = max_len

    def sentence_tokenizer(self, text: str, lower: bool = False) -> Generator[List[str], None, None]:
        sentence = []
        for token in text.split():
            sentence.append(token)
            if len(sentence) == self.max_len:
                yield sentence
                sentence = []
        if sentence:
            yield sentence

    def word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        if lower:
            return text.lower().split()
        return text.split()


from .processor import FrGlueProcessor
from pie_extended.pipeline.postprocessor.proto import RenamedTaskProcessor
from .tokenizer import FrMemorizingTokenizer
from pie_extended.pipeline.iterators.proto import DataIterator
from pie_extended.pipeline.postprocessor.memory import MemoryzingProcessor


def get_iterator_and_processor(max_tokens=256):
    tokenizer = FrMemorizingTokenizer()
    processor = MemoryzingProcessor(
            tokenizer_memory=tokenizer,
            head_processor=FrGlueProcessor(
                head_processor=RenamedTaskProcessor({"pos": "POS", "NOMB": "NOMB.", "PERS": "PERS."})
            )
        )

    iterator = DataIterator(
        tokenizer=tokenizer,
        max_tokens=max_tokens
    )
    return iterator, processor




import regex as re
from typing import List, Generator, Tuple

from pie_extended.pipeline.tokenizers.memorizing import MemorizingTokenizer
from pie_extended.pipeline.tokenizers.utils.excluder import ExcluderPrototype, DottedNumberExcluder, ApostropheExcluder, DEFAULT_CHAR_REGISTRY

_Dots_except_apostrophe = r".?!\"“”\"«»…\[\]\(\)„“"
_Dots_collections = r"[" + _Dots_except_apostrophe + "‘’]"


class FroMemorizingTokenizer(MemorizingTokenizer):
    re_add_space_around_punct = re.compile(r"(\s*)([^\w\s])(\s*)")
    re_remove_ending_apostrophe = re.compile(r"(?<=\w)([\'’ʼ])")
    _sentence_boundaries = re.compile(
        r"([" + _Dots_except_apostrophe + r"]+\s*)+"
    )

    def __init__(self):
        super(FroMemorizingTokenizer, self).__init__()
        self.tokens = []
        self.char_registry = DEFAULT_CHAR_REGISTRY
        self.excluders: Tuple[ExcluderPrototype, ...] = (
            DottedNumberExcluder(char_registry=DEFAULT_CHAR_REGISTRY),
            ApostropheExcluder()
        )

    @staticmethod
    def _sentence_tokenizer_merge_matches(match):
        """ Best way we found to deal with repeating groups"""
        start, end = match.span()
        return match.string[start:end] + "<SPLIT>"

    def _real_sentence_tokenizer(self, string: str) -> List[str]:
        string = self._sentence_boundaries.sub(self._sentence_tokenizer_merge_matches, string)

        for excluder in self.excluders:
            string = excluder.after_sentence_tokenizer(string)
        return string.split("<SPLIT>")

    def _real_word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        if lower:
            text = text.lower()
        text = text.split()
        return text

    def sentence_tokenizer(self, text: str, lower: bool = False) -> Generator[List[str], None, None]:
        sentences = list()
        data = self.normalizer(text)
        for sent in self._real_sentence_tokenizer(data):
            sent = sent.strip()
            if sent:
                sentences.append(self.word_tokenizer(sent))
        yield from sentences

    def normalizer(self, data: str) -> str:
        for excluder in self.excluders:
            data = excluder.before_sentence_tokenizer(data)

        data = self.re_add_space_around_punct.sub(
            r" \g<2> ",
            data
        )
        return data

    def replacer(self, inp: str):
        return self.re_remove_ending_apostrophe.sub("", inp)



import regex as re

from typing import List, Tuple, Dict, Iterable, Pattern, Union

from pie_extended.pipeline.tokenizers.simple_tokenizer import SimpleTokenizer
from enum import Enum


class GenericExcludePatterns(Enum):
    """ Useful set of regular expresion that can be used for the exclude_patterns

    """
    Punctuation_and_Underscore: Pattern = re.compile(r"^[_||[^\s\w]]+$", re.VERSION1)
    Punctuation: Pattern = re.compile(r"^[^\s\w]+$")
    PassageMarker: Pattern = re.compile(r"_Passage_[\w\d_]+")  # Use `_` as a joining character


class DataIterator:
    def __init__(self, tokenizer: SimpleTokenizer = None, exclude_patterns: List[Union[str, Pattern]] = None,
                 max_tokens: int = 256):
        """ Iterator used to parse the text and returns bits to tag

        :param tokenizer: Tokenizer
        """
        self.tokenizer: SimpleTokenizer = tokenizer or SimpleTokenizer()
        self.exclude_patterns: List[Pattern] = []
        if exclude_patterns:
            for pattern in exclude_patterns:
                self.add_pattern(pattern)
        self.max_tokens = max_tokens

    def add_pattern(self, pattern: str):
        """ Add a pattern for removal

        :param pattern: Pattern for token removal
        """
        if isinstance(pattern, str):
            self.exclude_patterns.append(re.compile(pattern))
        elif hasattr(pattern, "value"):  # Deal with enum
            self.exclude_patterns.append(pattern.value)
        else:
            self.exclude_patterns.append(pattern)

    def reset_patterns(self) -> None:
        """ Removes removal patterns

        >>> x = DataIterator(exclude_patterns=[r'\W+'])
        >>> x.exclude_tokens(["Je", "suis", "content", ",", "mais", "...", '"', "fatigué", '"', "."])
        (['Je', 'suis', 'content', 'mais', 'fatigué'], {3: ',', 5: '...', 6: '"', 8: '"', 9: '.'})
        >>> x.reset_patterns()
        >>> x.exclude_tokens(["Je", "suis", "content", ",", "mais", "...", '"', "fatigué", '"', "."])
        (['Je', 'suis', 'content', ',', 'mais', '...', '"', 'fatigué', '"', '.'], {})
        """
        self.exclude_patterns = []

    def exclude_tokens(self, sentence: List[str]) -> Tuple[List[str], Dict[int, str]]:
        """ Removes punctuation from a list and keeps its index

        :param sentence:
        :return: First the sentence with things removed, then a dictionary whose keys are index of token to reinsert and
        associated values are punctuation to reinsert.

        You can use string when generating the exclude_pattern

        >>> x = DataIterator(exclude_patterns=[r'\W+'])
        >>> x.exclude_tokens(["Je", "suis", "content",",", "mais", "...", '"', "fatigué", '"', "."])
        (['Je', 'suis', 'content', 'mais', 'fatigué'], {3: ',', 5: '...', 6: '"', 8: '"', 9: '.'})

        Pre-built removers:

        >>> x = DataIterator(exclude_patterns=[GenericExcludePatterns.PassageMarker])
        >>> x.exclude_tokens(["_Passage_45_78", "Ici", "commence", "le", "passage"])
        (['Ici', 'commence', 'le', 'passage'], {0: '_Passage_45_78'})

        And of course you can ignore this option

        >>> x = DataIterator()
        >>> x.exclude_tokens(["_Passage_45_78", "Ici", "commence", "le", "passage"])
        (['_Passage_45_78', 'Ici', 'commence', 'le', 'passage'], {})

        """
        if len(self.exclude_patterns) == 0:
            return sentence, {}

        clean, removed = [], {}
        for index, token in enumerate(sentence):
            match = False
            for exclude_pattern in self.exclude_patterns:
                if exclude_pattern.match(token):
                    removed[index] = token
                    match = True
                    break
            if not match:
                clean.append(token)

        return clean, removed

    def _max_out(self, sentences):
        for sentence in sentences:
            if len(sentence) <= self.max_tokens:
                yield sentence
            else:
                for n in range(0, len(sentence), self.max_tokens):
                    yield sentence[n:n+self.max_tokens]

    def __call__(self, data: str, lower: bool = False,
                 no_tokenizer: bool = False) -> Iterable[Tuple[List[str], int, Dict[int, str]]]:
        """ Default iter data takes a text, an option to make lower
        and yield lists of words along with the length of the list

        :param data: A plain text
        :param lower: Whether or not to lower the text
        :yields: (Sentence as a list of word, Size of the sentence, Elements removed from the sentence)
        """
        sentences = []

        func = self.tokenizer.sentence_tokenizer
        if no_tokenizer:
            func = self.tokenizer.bypass_tokenizer

        last_sentence_index = 0
        for sentence in self._max_out(func(data, lower=lower)):
            clean_sentence, removed_from_input = self.exclude_tokens(sentence)
            if len(clean_sentence) == 0 and len(sentences):
                sentences[-1][2].update({
                    last_sentence_index + removed_index: removed_value
                    for removed_index, removed_value in removed_from_input.items()
                })
                last_sentence_index += len(removed_from_input)
            else:
                sentences.append((clean_sentence, len(clean_sentence), removed_from_input))
                last_sentence_index = len(sentence)
        yield from sentences



from pie_extended.models.fro.imports import get_iterator_and_processor
from pie_extended.testing_utils import FakeTagger
from typing import List, Tuple

from unittest import TestCase
from .test_lasla import make_fake_data


def make_controller(sentences: List[str]):
    # Add the lemmatizer routes
    tagger = FakeTagger(
        make_fake_data(sentences),
        tasks="lemma,MODE,TEMPS,PERS,NOMB,GENRE,CAS,DEGRE,POS".split(",")
    )
    iterator, processor = get_iterator_and_processor()
    return tagger, iterator, processor


class TestFro(TestCase):
    def test_elision_apostrophe(self):
        string = "q'il meurt"
        treated = ["q il meurt"]
        tagger, it, pro = make_controller(treated)
        out = tagger.tag_str(string, it, pro)
        self.assertEqual(out[0]["form"], "q'")
        self.assertEqual(out[0]["treated"], "q")

    def test_elision_apostrophe_and_quote(self):
        string = "'q'il meurt 'dit il'"
        treated = ["q il meurt dit il"]
        tagger, it, pro = make_controller(treated)
        out = tagger.tag_str(string, it, pro)
        self.assertEqual(out[0]["form"], "'")
        self.assertEqual(out[0]["treated"], "'")
        self.assertEqual(out[1]["form"], "q'")
        self.assertEqual(out[1]["treated"], "q")
        self.assertEqual(out[-1]["form"], "'", "Last apostrophe is kept")
        # Ending and starting apostrophe are not reinserted for some reason.



import regex as re
from ...utils import get_path, ObjectCreator

from autocat import NeedsDisambiguation, StraightAutodisambiguation, CategoryAutodisambiguation, GroupAutodisambiguation
from ...pipeline.disambiguators.autocat import Autocat, DisambiguationAsTask
from pie_extended.models.lasla.processor import LatinRulesProcessor, LatinGlueProcessor, MoodTenseVoice
from pie_extended.pipeline.postprocessor.proto import ProcessorPrototype
from pie_extended.models.lasla.tokenizer import LatMemorizingTokenizer
from pie_extended.pipeline.iterators.proto import DataIterator, GenericExcludePatterns
from pie_extended.pipeline.postprocessor.memory import MemoryzingProcessor
from pie_extended.pipeline.postprocessor.splitter import SplitterPostProcessor

# Uppercase regexp
uppercase = re.compile(r"^[A-Z]$")


def get_iterator_and_processor(max_tokens=256):
    tokenizer = LatMemorizingTokenizer()
    processor = LatinRulesProcessor(
        apply_on_reinsert=True,
        head_processor=SplitterPostProcessor(
            prefix="界",
            split_char="界",
            head_processor=MemoryzingProcessor(
                tokenizer_memory=tokenizer,
                head_processor=LatinGlueProcessor(
                    head_processor=MoodTenseVoice(
                        head_processor=ProcessorPrototype()
                    )
                )
            )
        )
    )
    iterator = DataIterator(
        tokenizer=tokenizer,
        exclude_patterns=[
                             excl.exclude_regexp
                             for excl in tokenizer.normalizers
                             if excl.exclude_regexp
                         ] + [
                            GenericExcludePatterns.Punctuation_and_Underscore
                         ],
        max_tokens=max_tokens
    )
    return iterator, processor


def _get_disambiguator():
    pos = CategoryAutodisambiguation.from_file(
        get_path("lasla", "latin-pos.json"), category_key="pos", lemma_key="lemma")
    straight = StraightAutodisambiguation.from_file(
        get_path("lasla", "latin-straight.json"), lemma_key="lemma")
    impossible = NeedsDisambiguation.from_file(
        get_path("lasla", "latin-needs.json"), lemma_key="lemma")
    tasks = DisambiguationAsTask(task_name="Dis", separator="", undisambiguated_marker="?",
                                 lemma_key="lemma", empty_value="_", impossible_disambiguator=impossible)
    return Autocat(GroupAutodisambiguation(lemma_key="lemma", categorizers=(straight, pos, tasks)))


Disambiguator: ObjectCreator = ObjectCreator(_get_disambiguator)


def addons():
    return True



from pie_extended.pipeline.postprocessor.proto import ProcessorPrototype, ChainedProcessor
from typing import Optional, Dict, List
if "typing" == "nottyping":
    from ..tokenizers.memorizing import MemorizingTokenizer


class MemoryzingProcessor(ChainedProcessor):
    """ MemoryzingProcessor proposes to keep track of changes operated on input string
    by reinserting the original data alongside a new task (KEY) where we output
    the input seen by the Model

    It reuses the memory from a class derived from MemorizingTokenizer so that it reintroduced
    the original input into the token.

    >>> from pie_extended.pipeline.tokenizers.memorizing import MemorizingTokenizer
    >>> tokenizer = MemorizingTokenizer()
    >>> # Fake token memory : (Index, Original Input, Input seen by Tagger)
    >>> tokenizer.tokens = [(0, "A", "a"), (0, "b", "b"), (0, "q'", "q")]
    >>> processor = MemoryzingProcessor(tokenizer_memory=tokenizer, head_processor=ProcessorPrototype())
    >>> processor.set_tasks(["lem"])
    ['lem', 'treated']
    >>> # Lowercase a was taken in the input but uppercase a is returned in form. For transparency, input seen
    >>> #   By the tagger is returned in a new column, treated (cf. MemorizingProcessor.KEY)
    >>> processor.get_dict("a", ["lemma"]) == [{"form": "A", "treated": "a", "lem": "lemma"}]
    True
    >>> # Some would have the same treated and input
    >>> processor.get_dict("b", ["lemma"]) == [{"form": "b", "treated": "b", "lem": "lemma"}]
    True
    >>> # Some differ with more characters
    >>> processor.get_dict("q", ["lemma"]) == [{"form": "q'", "treated": "q", "lem": "lemma"}]
    True

    This allows for easier output alignment as well as removing unknown characters to the model. If your lemmatizer
    in training has never seen the "@" character, you can remove it at tokenization time and reinsert it with
    MemoryzingProcessor

    """
    KEY: str = "treated"

    def __init__(self, tokenizer_memory: "MemorizingTokenizer", head_processor: ProcessorPrototype,
                 key: Optional[str] = None, **kwargs):
        super(MemoryzingProcessor, self).__init__(head_processor=head_processor, **kwargs)
        self.memory: "MemorizingTokenizer" = tokenizer_memory
        self._key: str = key or type(self).KEY

    def get_dict(self, token: str, tags: List[str]) -> List[Dict[str, str]]:
        # First we get the dictionary
        list_token_dict = []
        for token_dict in self.head_processor.get_dict(token, tags):
            index, input_token, out_token = self.memory.tokens.pop(0)
            if token != out_token:
                raise Exception("The output token does not match our inputs %s : %s" % (token, out_token))

            token_dict[self._key] = out_token
            token_dict["form"] = input_token
            list_token_dict.append(token_dict)
        return list_token_dict

    @property
    def tasks(self) -> List[str]:
        return self.head_processor.tasks + ["treated"]

    def reinsert(self, form: str) -> Dict[str, str]:
        """ Reinsert the token, should add treated to the output

        >>> from pie_extended.pipeline.tokenizers.memorizing import MemorizingTokenizer
        >>> tokenizer = MemorizingTokenizer()
        >>> tokenizer.tokens = ["$"]
        >>> x = MemoryzingProcessor(tokenizer, ProcessorPrototype())
        >>> x.set_tasks(["task1", "task2"])
        ['task1', 'task2', 'treated']
        >>> x.reinsert("$")
        {'form': '$', 'task1': '_', 'task2': '_', 'treated': '--IGN.--'}
        """
        self.memory.tokens.pop(0)
        annotations = super(MemoryzingProcessor, self).reinsert(form)
        annotations['treated'] = '--IGN.--'
        return annotations



from collections import namedtuple
import os

Metadata = namedtuple("Metadata", ["title", "lang", "authors", "description", "link"])

_File = namedtuple("File", ["url", "name"])


class File(_File):
    def __str__(self):
        return self.name


PATH = os.getenv(
    "PIE_EXTENDED_DOWNLOADS",
    os.path.normpath(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..",
            "downloads"
        )
    )
)


def get_path(module, file):
    return os.path.join(PATH, module, file)


class ObjectCreator:
    """ Some objects should be reset everytime a new tagging is done. To make this easier
    we provide this class that keeps in memory the initialization parameters."""
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def create(self):
        return self.cls(*self.args, **self.kwargs)


def roman_number(inp: str) -> int:
    """
    Source: https://stackoverflow.com/questions/19308177/converting-roman-numerals-to-integers-in-python
    Author: https://stackoverflow.com/users/1201737/r366y
    :param num:
    :return:

    >>> roman_number("XXIV")
    24
    """
    roman_numerals = {'I': 1, 'V': 5, 'X': 10, 'L': 50, 'C': 100, 'D': 500, 'M': 1000}
    result = 0
    for i, c in enumerate(inp.upper()):
        if (i+1) == len(inp) or roman_numerals[c] >= roman_numerals[inp[i+1]]:
            result += roman_numerals[c]
        else:
            result -= roman_numerals[c]
    return result



from ...utils import Metadata, File, get_path


DESC = Metadata(
    "LASLA-ENC",
    "lat",
    ["Thibault Clérice"],
    "Model trained on LASLA data without disambiguation",
    "https://github.com/chartes/deucalion-model-lasla"
)

VERSION = "0.0.5b"
DOWNLOADS = [
    File("https://raw.githubusercontent.com/PonteIneptique/latin-lasla-models/"+VERSION+"/latin-straight.json",
         "latin-straight.json"),
    File("https://raw.githubusercontent.com/PonteIneptique/latin-lasla-models/"+VERSION+"/latin-pos.json",
         "latin-pos.json"),
    File("https://raw.githubusercontent.com/PonteIneptique/latin-lasla-models/"+VERSION+"/latin-needs.json",
         "latin-needs.json"),
    File("https://github.com/PonteIneptique/latin-lasla-models/releases/download/"+VERSION+"/lasla-plus.tar",
         "model.tar")
]

Models = "<{},lemma,Deg,Numb,Person,Mood_Tense_Voice,Case,Gend,Dis,pos>".format(get_path("lasla", "model.tar"))



from ...utils import Metadata, File, get_path

DESC = Metadata(
    "Ancien Francais (ENC)",
    "fro",
    ["Jean-Baptiste Camps", "Ariane Pinche", "Thibault Clérice", "Frédéric Duval", "Lucence Ing", "Naomi KANAOKA"],
    "Model trained on diverse corpora (hagiography, geste, etc.)",
    "https://github.com/chartes/deucalion-model-af"
)

VERSION = "0.3.0"

DOWNLOADS = [
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/lemma-pos.tar",
         "lemma-pos.tar"),
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/cas.tar",
         "cas.tar"),
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/degre.tar",
         "degre.tar"),
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/genre.tar",
         "genre.tar"),
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/mode.tar",
         "mode.tar"),
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/nomb.tar",
         "nomb.tar"),
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/pers.tar",
         "pers.tar"),
    File("https://github.com/chartes/deucalion-model-af/releases/download/" + VERSION + "/temps.tar",
         "temps.tar"),
]

Models = "".join([
    "<{},MODE>".format(get_path("fro", "MODE.tar".lower())),
    "<{},TEMPS>".format(get_path("fro", "TEMPS.tar".lower())),
    "<{},PERS>".format(get_path("fro", "PERS.tar".lower())),
    "<{},NOMB>".format(get_path("fro", "NOMB.tar".lower())),
    "<{},GENRE>".format(get_path("fro", "GENRE.tar".lower())),
    "<{},CAS>".format(get_path("fro", "CAS.tar".lower())),
    "<{},DEGRE>".format(get_path("fro", "DEGRE.tar".lower())),
    "<{},lemma,POS>".format(
        get_path("fro", "lemma-pos.tar")
    )
])



from .simple_tokenizer import SimpleTokenizer, RE_BYPASS_SENTENCE, RE_BYPASS_WORD
from typing import List, Tuple, Dict, Generator


class MemorizingTokenizer(SimpleTokenizer):
    """ Tokenizer that memoryze what it tokenized.

    Mostly used to normalized input as input time and then reinserting normalized input

    """

    def replacer(self, token: str) -> str:
        """ This function allows for changing input and keeping it in memory """
        return token

    def __init__(self):
        self.tokens: List[Tuple[int, str, str]] = []

    def _real_word_tokenizer(self, data: str, lower: bool = False) -> List[str]:
        return super(MemorizingTokenizer, self).word_tokenizer(data, lower=lower)

    def _word_logic(self, token: str) -> str:
        out = self.replacer(token)
        self.tokens.append((len(self.tokens), token, out))
        return out

    def word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        sentence = []
        for token in self._real_word_tokenizer(text, lower):
            sentence.append(self._word_logic(token))
        return sentence

    def reset(self):  # Empty
        self.tokens = []

    def bypass_tokenizer(self, data: str, lower: bool = False) -> Generator[List[str], None, None]:
        """ Function to enable pretokenized input while using replaces or the likes

       :param data: The string (one new line = one word, two new lines = two words)
       :param lower: Whether to lower the input


       >>> tokenizer = MemorizingTokenizer()
       >>> list(tokenizer.bypass_tokenizer("One\\ntwo\\nthree\\n\\n.//.\\na\\nz"))
       [['One', 'two', 'three'], ['.//.', 'a', 'z']]
       >>> tokenizer.tokens
       [(0, 'One', 'One'), (1, 'two', 'two'), (2, 'three', 'three'), (3, './/.', './/.'), (4, 'a', 'a'), (5, 'z', 'z')]

       """
        for sentence in RE_BYPASS_SENTENCE.split(data):
            if sentence:
                yield [self._word_logic(word) for word in RE_BYPASS_WORD.split(sentence)]






from typing import Dict, Generator
from pie_extended.pipeline.postprocessor.glue import GlueProcessor


class FrGlueProcessor(GlueProcessor):
    """ We glue morphological features into one column
    """
    OUTPUT_KEYS = ["form", "lemma", "POS", "morph"]
    GLUE = {"morph": ["MODE", "TEMPS", "PERS.", "NOMB.", "GENRE", "CAS"]}  # , "DEGRE"]}
    EMPTY_TAG: Dict[str, str] = {"CAS": "CAS=x", "NOMB.": "NOMB.=x", "MODE": "MODE=x", "TEMPS": "TEMPS=x",
                                 "GENRE": "GENRE=x", "PERS.": "PERS.=x"}

    # We need to change the definition of this function, because, for now,
    # the model gives us an output like
    # MODE=imp
    # and not
    # imp
    def _yield_annotation(
            self,
            token_dict: Dict[str, str]
        ) -> Generator[str, None, None]:
        # For each key we should return
        for head in self._out:
            if head not in self._glue:
                yield head, token_dict[head]
            else:
                # Otherwise, we glue together things that should be glued together
                joined = self._glue_char.join([
                    # HERE
                    # glued_task + "=" +
                    token_dict[glued_task]
                    for glued_task in self._glue[head]
                    if token_dict[glued_task] != self._empty_tags.get(glued_task, None)
                ])
                if not joined:
                    joined = self._glue_empty[head]
                yield head, joined

    def __init__(self, *args, **kwargs):
        super(FrGlueProcessor, self).__init__(*args, **kwargs)



from typing import List

LASLA = ['A', 'Ap', 'App', 'Apr', 'April', 'C', 'Cn', 'D', 'Epigr', 'Febr', 'G', 'III', 'Ian', 'Id', 'Iun', 'Kal',
         'Kalend', 'L', 'LVIIII', 'LXXII', 'LXXXIIII', 'M', 'Mam', 'Mart', 'N', 'Non', 'Nou', 'Octob', 'P', 'Q',
         'Quintil', 'R', 'S', 'Ser', 'Seru', 'Sex', 'Sp', 'T', 'Ti', 'Tib', 'VII', 'XIII']


PRAENOMINA = [
    "a",
    "agr",
    "ap",
    "c",
    "cn",
    "d",
    "f",
    "k",
    "l",
    "m'",
    "m",
    "mam",
    "n",
    "oct",
    "opet",
    "p",
    "post",
    "pro",
    "q",
    "s",
    "ser",
    "sert",
    "sex",
    "st",
    "t",
    "ti",
    "v",
    "vol",
    "vop",
    "a",
    "ap",
    "c",
    "cn",
    "d",
    "f",
    "k",
    "l",
    "m",
    "m'",
    "mam",
    "n",
    "oct",
    "opet",
    "p",
    "paul",
    "post",
    "pro",
    "q",
    "ser",
    "sert",
    "sex",
    "sp",
    "st",
    "sta",
    "t",
    "ti",
    "v",
    "vol",
    "vop",
]

CALENDAR = [
    "ian",
    "febr",
    "mart",
    "apr",
    "mai",
    "iun",
    "iul",
    "aug",
    "sept",
    "oct",
    "nov",
    "dec",
] + ["kal", "non", "id", "a.d"]

MISC = ["coll", "cos", "ord", "pl.", "s.c", "suff", "trib"]

CLTK = [
    string.replace("v", "u").replace("j", "i").capitalize()+"."
    for string in MISC + CALENDAR + PRAENOMINA
]

abbrs: List[str] = sorted(list(set(CLTK + LASLA)))



from pie_extended.pipeline.postprocessor.proto import ChainedProcessor, ProcessorPrototype, RenamedTaskProcessor
from typing import Generator, Dict, List


class GlueProcessor(ChainedProcessor):
    """ Glues together specific tasks

    >>> class SimpleGlue(GlueProcessor):
    ...     OUTPUT_KEYS = ["form", "lemma", "task3"]
    ...     GLUE = {"task3": ["1", "2"]} # Merges Task `1` output and task `2` output in `task3`
    ...     EMPTY_TAG = {"1": "_", "2": "_"} # If _ is tagged in task `1`, it's the same as an empty tag
    ...     GLUE_EMPTY = {"task3": "NO-DATA"}  # When all merged data are empty, default value
    >>> x = SimpleGlue(head_processor=ProcessorPrototype())
    >>> x.set_tasks(["lemma", "1", "2"]) # You can see things are remaped
    ['lemma', 'task3']
    >>> # Merges b and c values from task 1 and 2 into a new task
    >>> x.get_dict("a", ["a", "b", "c"]) == [{"form": "a", "lemma": "a", "task3": "1=b|2=c"}]
    True
    >>> # Keeps only one task because 2 is empty
    >>> x.get_dict("a", ["a", "b", "_"]) == [{"form": "a", "lemma": "a", "task3": "1=b"}]
    True
    >>> # Fills with the default empty tag because both task 1 and 2 were empty
    >>> x.get_dict("a", ["a", "_", "_"]) == [{"form": "a", "lemma": "a", "task3": "NO-DATA"}]
    True

    You can also use remaped tasks:

    >>> class AnotherGlue(GlueProcessor):
    ...     OUTPUT_KEYS = ["form", "lemma", "POS", "task3"]
    ...     GLUE = {"task3": ["1", "2"]} # Merges Task `1` output and task `2` output in `task3`
    ...     EMPTY_TAG = {"1": "_", "2": "_"} # If _ is tagged in task `1`, it's the same as an empty tag
    ...     GLUE_EMPTY = {"task3": "NO-DATA"}  # When all merged data are empty, default value
    >>> x = AnotherGlue(head_processor=RenamedTaskProcessor({"pos": "POS"}))
    >>> x.set_tasks(["lemma", "pos", "1", "2"]) # You can see things are remaped
    ['lemma', 'POS', 'task3']
    >>> # Merges b and c values from task 1 and 2 into a new task
    >>> x.get_dict("a", ["a", "p", "b", "c"])
    [{'form': 'a', 'lemma': 'a', 'POS': 'p', 'task3': '1=b|2=c'}]

    """

    # Output keys are keys that are given in the end
    OUTPUT_KEYS: List[str] = ["form", "lemma", "POS", "morph"]
    # Glue dicts contains tasks that should merge together subtasks
    GLUE: Dict[str, List[str]] = {"morph": ["Case", "Numb", "Deg", "Mood", "Tense", "Voice", "Person"]}
    # Glue_char is what is used to glue things together -> Tense=Pres|Person=1
    GLUE_CHAR: str = "|"
    # Glue Empty are value to take when all things glued together are empty
    GLUE_EMPTY: Dict[str, str] = {"morph": "MORPH=empty"}
    # Value that means the current element is empty
    EMPTY_TAG: Dict[str, str] = {"Case": "_", "Numb": "_", "Deg": "_", "Mood": "_", "Tense": "_", "Voice": "_",
                                 "Person": "_"}

    def __init__(self, *args, **kwargs):
        super(GlueProcessor, self).__init__(*args, **kwargs)

        # Sets-up some copy of the values
        self._out = self.OUTPUT_KEYS
        self._glue = self.GLUE
        self._glue_char = self.GLUE_CHAR
        self._glue_empty = self.GLUE_EMPTY
        self._empty_tags = self.EMPTY_TAG

    def _yield_annotation(
            self, 
            token_dict: Dict[str, str]
        ) -> Generator[str, None, None]:
        # For each key we should return
        for head in self._out:
            if head not in self._glue:
                yield head, token_dict[head]
            else:
                # Otherwise, we glue together things that should be glued together
                joined = self._glue_char.join([
                    self._get_glued(glued_task, token_dict)
                    for glued_task in self._glue[head]
                    if token_dict[glued_task] != self._empty_tags.get(glued_task, None)
                ])
                if not joined:
                    joined = self._glue_empty[head]
                yield head, joined

    def _get_glued(self, glued_task: str, token_dict: Dict[str, str]):
        return glued_task + "=" + token_dict[glued_task]

    def reinsert(self, form: str) -> Dict[str, str]:
        return dict(form=form, **{key: self.empty_value for key in self._out if key != "form"})

    def get_dict(self, token: str, tags: List[str]) -> List[Dict[str, str]]:
        return [dict(self._yield_annotation(as_dict)) for as_dict in super(GlueProcessor, self).get_dict(token, tags)]

    @property
    def tasks(self) -> List[str]:
        return [key for key in self._out if key != "form"]



from typing import List, Iterable, Callable, Dict
import sys


class Formatter:  # Default is TSV
    """ The CSV formatter necessarily starts with form in its header.

    """
    format_line: Callable[[Dict[str, str]], List[str]]

    def __init__(self, tasks: List[str]):
        self.tasks: List[str] = tasks

        # Before 3.7, order of dictionary is not guaranteed
        # Cf. https://mail.python.org/pipermail/python-dev/2017-December/151283.html
        self.format_line = self.format_line_3_6
        # With post-processing, it's better to not trust order
        #else:
        #    self.format_line = self.format_line_3_7

    def format_line_3_6(self, annotation: Dict[str, str]) -> List[str]:
        """ Format the tags """
        return [annotation["form"]] + [annotation[task] for task in self.tasks]

    def format_line_3_7(self, annotation: Dict[str, str]) -> List[str]:
        """ Format the tags """
        return list(annotation.values())

    def write_line(self, formatted):
        return "\t".join(formatted) + "\r\n"

    def write_sentence_beginning(self) -> str:
        return ""

    def write_sentence_end(self) -> str:
        return ""

    def write_footer(self) -> str:
        return ""

    def get_headers(self):
        return ["token"] + self.tasks

    def write_headers(self)-> str:
        """ Format the headers """
        return self.write_line(self.get_headers())



from typing import List, Tuple, Optional
from pie_extended.pipeline.iterators.proto import DataIterator
from pie_extended.pipeline.postprocessor.proto import ProcessorPrototype
from pie_extended.tagger import ExtensibleTagger
from pie.utils import model_spec
from pie_extended.cli.utils import get_imports


class FakeTagger(ExtensibleTagger):
    def __init__(self, tokens, tasks):
        self.tokens = tokens
        self.tasks = tasks
        self.seen = []
        self.lower = False
        self.batch_size = 100
        self.disambiguation = None

    def tag(self, sents, **kwargs):
        self.seen.extend(sents)

        return self.tokens, self.tasks


class FakeAutoTag(ExtensibleTagger):
    def __init__(self, tasks: List[str], **kwargs):
        self.tokens: List[str] = []
        self.lengths: List[int] = []
        self.tasks = tasks
        self.lower = False
        self.glue_task: str = kwargs.get("glue_task", "_")
        self.glue_value: str = kwargs.get("glue_value", "|")
        self.disambiguation = None
        self.start_end: bool = False
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def toggle_start_end(self, status: Optional[bool] = None) -> bool:
        if status is None:
            self.start_end = not self.start_end
        else:
            self.start_end = status
        return self.start_end

    def tag(self, sents: List[List[str]], lengths: List[int], *args, **kwargs):
        """ Fake tagging tokens by enumerating informations

        >>> tagger = FakeAutoTag(["pos", "lemma"])
        >>> tagger.tag([['a', 'b'], ['c']], lengths=[2, 1])
        ([[('a', ('pos0', 'lemma0')), ('b', ('pos1', 'lemma1'))], [('c', ('pos2', 'lemma2'))]], ['pos', 'lemma'])

        """
        self.tokens.extend(list(sents))
        self.lengths.extend(lengths)

        for t, l in zip(sents, lengths):
            if len(t) != l:
                raise ValueError("Tokens and lengths are inequal [len({}) != {}]".format(str(t), l))

        out = []
        total = 0

        def get_task(task, i):
            if "_" in task:
                return self.glue_value.join([tsk+str(i) for tsk in task.split(self.glue_task)])
            return task+str(i)

        for sent in sents:
            out.append([])
            if self.start_end:
                out[-1].append(("START", tuple(list(get_task(task, total) for task in self.tasks))))
            for tok in sent:
                out[-1].append((tok, tuple(list(get_task(task, total) for task in self.tasks))))
                total += 1
            if self.start_end:
                out[-1].append(("END", tuple(list(get_task(task, total) for task in self.tasks))))
        return out, self.tasks

    @staticmethod
    def from_model_string(model_string: str, **kwargs) -> "FakeAutoTag":
        """

        :param model_string:
        :return:

        >>> tagger = FakeAutoTag.from_model_string("<path/to/tar,MODE,TEMPS,PERS,NOMB><path/to/tar,lemma,pos>")
        >>> tagger.tasks
        ['MODE', 'TEMPS', 'PERS', 'NOMB', 'lemma', 'pos']
        """
        return FakeAutoTag(tasks=[
            task
            for _, tasks in model_spec(model_string)
            for task in tasks
        ], **kwargs)


def create_auto_tagger(module, **kwargs) -> Tuple[FakeAutoTag, DataIterator, ProcessorPrototype]:
    """ Create a tagger as well as the iterator """
    tagger = FakeAutoTag.from_model_string(module.Models, batch_size=16, **kwargs)
    imports = get_imports(module)
    disambiguator = getattr(imports, "Disambiguator", None)
    if hasattr(disambiguator, "create"):
        disambiguator = disambiguator.create()
    tagger.disambiguation = disambiguator

    iterator, processor = imports.get_iterator_and_processor()
    return tagger, iterator, processor



from autocat.autocat import ProtoDisambiguator, NeedsDisambiguation
from .proto import Disambiguator
from typing import Optional, Set


class Autocat(Disambiguator):
    def __init__(self, autocat_categorizer: ProtoDisambiguator, separator: str = "", **kwargs):
        self.autocat_categorizer: ProtoDisambiguator = autocat_categorizer
        self.lemma_key = autocat_categorizer.lemma_key
        self.separator = separator

    def __call__(self, sent, tasks):
        toks, sents_as_rows = zip(*[
            (token, dict(zip(tasks, tags)))
            for token, tags in sent
        ])

        out = [
        ]

        for (tok, tags, dis) in zip(toks, sents_as_rows, self.autocat_categorizer.disambiguate_rows(sents_as_rows)):
            if dis:
                tags[self.lemma_key] += dis
            out.append((tok, list(tags.values())))

        return out


class DisambiguationAsTask(ProtoDisambiguator):
    def __init__(self,
                 impossible_disambiguator: NeedsDisambiguation,
                 task_name: str = "Dis", separator: str = "",
                 undisambiguated_marker: str = "?",
                 lemma_key: str = "lemma", empty_value: str = "_"
                 ):
        self.task_name: str = task_name
        self.undisambiguated_marker: str = undisambiguated_marker
        self.separator: str = separator
        self.lemma_key: str = lemma_key
        self.empty_value: str = empty_value
        self.impossible_disambiguator: NeedsDisambiguation = impossible_disambiguator

    @property
    def known_tokens(self) -> Set[str]:
        return self.impossible_disambiguator.known_tokens

    def disambiguate(self, tasks) -> Optional[str]:
        disamb = self.impossible_disambiguator.disambiguate(tasks)
        if disamb and tasks[self.task_name] != self.undisambiguated_marker:
            disamb = tasks[self.task_name]
            if disamb == self.empty_value:
                return self.undisambiguated_marker
            return disamb
        else:
            return disamb

    @classmethod
    def from_file(cls, filepath, lemma_key="", **kwargs):
        return None



from abc import ABC
import regex as re
from typing import Match, List, Optional, Dict

import pie_extended.pipeline.tokenizers.utils.chars as chars
import pie_extended.pipeline.tokenizers.utils.regexps as regexps
import copy

# Common values so that there is not (too much) collision
DOT = '語'
COLON = '桁'
BRACKET_L = '左'
BRACKET_R = '右'
APOSTROPHE = '風'
DASH = "精"


_DEFAULT_CHAR_REGISTRY = {
    ".": '語',
    ":": '桁',
    '[': '左',
    ']': '右',
    "'": '風',
    "-": "精"
}


class CharRegistry:
    def __init__(self, use_default=True, unicode_start_range=0x8d00):
        self._char_to_code = {}
        if use_default:
            self._char_to_code.update(_DEFAULT_CHAR_REGISTRY)
        self._start_range = unicode_start_range

    def __getitem__(self, item):
        if item in self._char_to_code:
            return self._char_to_code[item]
        else:
            self._char_to_code[item] = chr(self._start_range + len(self._char_to_code))
            return self._char_to_code[item]

    def __setitem__(self, key, value):
        self._char_to_code[key] = value

    def items(self):
        return self._char_to_code.items()

    def __len__(self):
        return len(self._char_to_code)


DEFAULT_CHAR_REGISTRY = CharRegistry()


class ExcluderPrototype(ABC):
    """Excluders should be used to remove dots and the likes from string before sentence tokenization.
    Data can be reintroduce later with their .denormalize() method"""
    def before_sentence_tokenizer(self, value: str) -> str:
        raise NotImplementedError

    def after_sentence_tokenizer(self, value: str) -> str:
        raise NotImplementedError

    @property
    def can_be_replaced(self) -> bool:
        return True

    @property
    def exclude_regexp(self) -> Optional[re.Regex]:
        return None


class ReferenceExcluder(ExcluderPrototype):
    """ Allows for exclusion of reference tokens such as [REF:1.b.Z]

    >>> ref = ReferenceExcluder()
    >>> ref.before_sentence_tokenizer("ici [REF:###abc??<>] Paris")
    'ici  左REF桁贆贆贆abc贉贉贇贈右  Paris'
    >>> ref.after_sentence_tokenizer('ici  左REF桁贆贆贆abc贉贉贇贈右  Paris')
    'ici  [REF:###abc??<>]  Paris'
     """

    def __init__(self,
                 regex_string: str = r"(\[REF:[^\]]+\])",
                 regex_needs_replacement: str = r"[^\w\s]",
                 char_registry: Optional[CharRegistry] = None
                 ):
        self.re: re.Regex = re.compile(regex_string)
        self.char_registry: CharRegistry = char_registry or CharRegistry()
        # Character that needs to be escaped.
        self.needs_replacement: re.Regex = re.compile(regex_needs_replacement)

    @property
    def can_be_replaced(self) -> bool:
        return False

    @property
    def exclude_regexp(self) -> Optional[re.Regex]:
        return self.re

    def _replace_in(self, match: Match) -> str:
        data = match.group()
        for char, repl in self.char_registry.items():
            data = data.replace(char, repl)

        for char in sorted(list(set(self.needs_replacement.findall(data)))):
            data = data.replace(char, self.char_registry[char])

        return f" {data} "

    def before_sentence_tokenizer(self, value: str) -> str:
        """ Normalize a string before it goes into sentence tokenizing

        :param value: String to clean

        >>> excl = ReferenceExcluder()
        >>> excl.before_sentence_tokenizer("Choubidou [REF:1.a.Z] choubida [REF:1.b.Z]")
        'Choubidou  左REF桁1語a語Z右  choubida  左REF桁1語b語Z右 '

        """
        return self.re.sub(self._replace_in, value)

    def after_sentence_tokenizer(self, value: str) -> str:
        """ Reset the state of a string before it goes into sentence tokenizing

        :param value: String to clean

        >>> excl = ReferenceExcluder()
        >>> excl.after_sentence_tokenizer('Choubidou 左REF桁1語a語Z右 choubida 左REF桁1語b語Z右')
        'Choubidou [REF:1.a.Z] choubida [REF:1.b.Z]'

        """
        for char, repl in self.char_registry.items():
            value = value.replace(repl, char)
        return value

    def ignore(self, string: str) -> bool:
        return bool(self.re.match(string))


class RegexpExcluder(ExcluderPrototype):
    def __init__(self, regex: str):
        """

        :param regex: Regular expression to exclude from tokenization

        Exclude data from tokenization

        >>> r = RegexpExcluder(r"(\p{No})")
        >>> r.can_be_replaced
        False
        >>> r.exclude_regexp.match("¹") is None
        False
        >>> r.exclude_regexp.match("2") is None
        True
        """
        self.re = re.compile(regex)

    @property
    def can_be_replaced(self) -> bool:
        return False

    @property
    def exclude_regexp(self) -> Optional[re.Regex]:
        return self.re

    def before_sentence_tokenizer(self, value: str) -> str:
        return value

    def after_sentence_tokenizer(self, value: str) -> str:
        return value


class AbbreviationsExcluder(ExcluderPrototype):
    def __init__(self, abbrs: List[str], apply_replacements: bool = True,
                 char_registry: Optional[CharRegistry] = None):
        """

        :param: List of abbreviation (dot included), eg. ['cf.', 'p.']
        """
        self.char_registry: CharRegistry = char_registry or CharRegistry()
        self.re = re.compile(
            r"("+r"|".join([abbr.replace(".", "") for abbr in abbrs])+r")(\.)"
        )
        self._apply_replacements = apply_replacements

    def _replace_in(self, match: Match) -> str:
        return match.group()\
            .replace(".", self.char_registry["."])

    @property
    def can_be_replaced(self) -> bool:
        return self._apply_replacements

    @property
    def exclude_regexp(self) -> Optional[re.Regex]:
        return self.re

    def before_sentence_tokenizer(self, value: str) -> str:
        """ Normalize a string before it goes into sentence tokenizing

        :param value: String to clean

        >>> excl = AbbreviationsExcluder(['cf.', 'p.'])
        >>> excl.before_sentence_tokenizer("cf. p. 45 et les. points.")
        'cf語 p語 45 et les. points.'

        """
        return self.re.sub(self._replace_in, value)

    def after_sentence_tokenizer(self, value: str) -> str:
        """ Reset the state of a string before it goes into sentence tokenizing

        :param value: String to clean

        >>> excl = AbbreviationsExcluder(['cf.', 'p.'])
        >>> excl.after_sentence_tokenizer('cf語 p語 45 et les. points.')
        'cf. p. 45 et les. points.'

        """
        return value.replace(
            self.char_registry["."], '.'
        )


class AbbreviationsRemoverExcluder(AbbreviationsExcluder):
    def after_sentence_tokenizer(self, value: str) -> str:
        """ Reset the state of a string before it goes into sentence tokenizing

        :param value: String to clean

        >>> excl = AbbreviationsRemoverExcluder(['cf.', 'p.'])
        >>> excl.after_sentence_tokenizer('cf語 p語 45 et les. points.')
        'cf p 45 et les. points.'

        """
        return value.replace(self.char_registry["."], '')


class CompoundAbbreviationsExcluder(AbbreviationsExcluder):
    def __init__(self, abbrs: List[str], apply_replacements: bool = True,
                 ignore_case: bool = True, char_registry: Optional[CharRegistry] = None):
        """

        >>> excl = CompoundAbbreviationsExcluder(["cf.", "V. act."])
        >>> excl.before_sentence_tokenizer("Cf. article 5, V. act. 9.")
        'Cf語 article 5, V語 act語 9.'
        >>> excl.after_sentence_tokenizer('Cf語 article 5, V語 act語 9.')
        'Cf. article 5, V. act. 9.'
        """
        super(CompoundAbbreviationsExcluder, self).__init__(abbrs, apply_replacements, char_registry=char_registry)
        re_kwargs = {}
        if ignore_case:
            re_kwargs["flags"] = re.IGNORECASE
        self.re = re.compile(
            r"\b("+"|".join([
                token.replace(" ", r"\s+").replace(".", r"\.")
                for token in abbrs
                if token  # Small check.
            ])+r")",
            **re_kwargs
        )


class DottedNumberExcluder(ExcluderPrototype):
    def __init__(self, number_regex: str = regexps.RomanNumbers,
                 char_registry: Optional[CharRegistry] = None):
        self.re = re.compile(r"\.(" + number_regex + r")\.")
        self.char_registry: CharRegistry = char_registry or CharRegistry()

    def before_sentence_tokenizer(self, value: str) -> str:
        return self.re.sub(
            r"{0}\g<1>{0}".format(self.char_registry["."]),
            value,
        )

    def after_sentence_tokenizer(self, value: str) -> str:
        return value.replace(
            self.char_registry["."], '.'
        )


class ApostropheExcluder(ExcluderPrototype):
    """Allows for keeping apostrophes that are the mark of an elision, mostly in French such as
    "l'abbé"
    """
    def __init__(self,
                 match_apostrophes=chars.APOSTROPHE,
                 apostrophe_mask=APOSTROPHE,
                 add_space_after: bool = True,
                 add_space_before: bool = False):
        self.apostrophes = match_apostrophes
        self.re: re.Regex = re.compile(r"(\w+)([" + self.apostrophes + r"])(\w+)")
        self.apostrophe_mask: str = apostrophe_mask

        # Space handling
        self.space_before: str = ""
        if add_space_before:
            self.space_before = " "

        self.space_after: str = ""
        if add_space_after:
            self.space_after = " "

    def _before_sentence_tokenizer(self, regex_match: Match) -> str:
        """ Given a match on `l'abbé` returns `l風0abbé` where
            風 is the character mask and 0 the index of the apostroph in the match_apostrophes value
        """
        return regex_match.group(1) + \
            self.apostrophe_mask + \
            str(self.apostrophes.index(regex_match.group(2))) + \
            regex_match.group(3)

    def before_sentence_tokenizer(self, value: str) -> str:
        """ Normalization run before sentence tokenization

        :param value: String to normalize before sentence is toknized

        >>> excl = ApostropheExcluder(add_space_after=True)
        >>> excl.before_sentence_tokenizer("l'abbé lorsqu’il")
        'l風0abbé lorsqu風1il'

        """
        return self.re.sub(
            self._before_sentence_tokenizer,
            value
        )

    def after_sentence_tokenizer(self, value: str) -> str:
        """ Normalization run after sentence tokenization

        :param value: String to normalize before sentence is toknized

        >>> excl = ApostropheExcluder(add_space_after=True)
        >>> excl.after_sentence_tokenizer('l風0abbé lorsqu風1il')
        "l' abbé lorsqu’ il"

        """
        for index_apo, apostrophe in enumerate(self.apostrophes):
            value = value.replace(
                self.apostrophe_mask+str(index_apo),
                self.space_before + apostrophe + self.space_after)
        return value



from pie_extended.models.lasla.imports import get_iterator_and_processor
from pie_extended.models import lasla
from pie_extended.testing_utils import FakeTagger, create_auto_tagger
from typing import List, Tuple

from unittest import TestCase


def write_crazy_file() -> str:
    filename = "crazy_text_file.txt"
    with open(filename, "w") as f:
        f.write("""\\\\\\<1>[$@$](V)\\\\\\§
\\\\<1>[$@$]\\\\§
§
\\[I]\\§
At o sceleste penis, o meum malum,
graui piaque lege noxiam lues.
Licet querare, nec tibi tener puer""")
    return filename


class FakeAnnotationMaker:
    def __init__(self):
        self.cnt = 0

    def make_response(self, token, is_ignored=False, treated=None):
        if is_ignored:
            return {
                'token': token, 'lemma': token, 'pos': 'PUNC', 'morph': 'MORPH=empty', 'Dis': '_',
                'treated': '--IGN.--'
            }
        else:
            resp = {
                'token': token, 'lemma': f'lemma{self.cnt}', 'pos': f'pos{self.cnt}',
                'morph': 'Case=Case{0}|Numb=Numb{0}|Gend=Gend{0}|Deg=Deg{0}|Mood=Mood{0}|Tense=Tense{0}|'
                         'Voice=Voice{0}|Person=Person{0}'.format(self.cnt),
                'Dis': f'Dis{self.cnt}', 'treated': treated if treated else token
            }
            self.cnt += 1
            return resp

    def make_multiple_response(self, tokens, is_ignored=False):
        for token in tokens:
            yield self.make_response(token, is_ignored=is_ignored)

def make_fake_data(sentences: List[str], nb_tasks: int = 9) -> List[Tuple[str, List[str]]]:
    return [
        [
            (token, tuple([token, *["fake"]*nb_tasks]))
            for token in sentence.split()
        ]
        for sentence in sentences
    ]


def make_controller(sentences: List[str]):
    # Add the lemmatizer routes
    tasks = "lemma,Voice,Mood,Deg,Numb,Person,Tense,Case,Gend,Dis,pos".split(",")
    tagger = FakeTagger(
        make_fake_data(sentences, nb_tasks=len(tasks)),
        tasks=tasks
    )
    iterator, processor = get_iterator_and_processor()
    return tagger, iterator, processor


class TestLasla(TestCase):
    def test_consecutive_dots(self):
        """Check that consecutive punctation does not break anything

        Found out the hard way it would break things
        """

        tagger, data_iterator, processor = make_controller([
            "id enim ait turbabuntur a facie eius patris or phanorum et iudicis uiduarum",
            "causam turbationis hanc docuit quod pater"
        ])

        result = tagger.tag_str(
            data="id enim ait turbabuntur a facie eius patris or phanorum et iudicis uiduarum ."
                          "  .  causam turbationis hanc docuit quod pater",
            processor=processor,
            iterator=data_iterator
        )
        self.assertEqual(
            result[12],
            {"form": "uiduarum", "lemma": "uiduarum", "pos": "fake", "morph": "Case=fake|Numb=fake|Gend=fake"
                                                                              "|Deg=fake|Mood=fake|"
                                                                              "Tense=fake|Voice=fake|Person=fake",
                                                                              'Dis': 'fake',
             "treated": "uiduarum"},
            "Punctuation should be reinserted and mostly should not break anything"
        )

    def test_leading_punctuation(self):
        """Check that consecutive punctation does not break anything

        Special case of consecutive dots, where sentences starts with it
        """
        tagger, data_iterator, processor = make_controller([
            # Need an empty sentence because ( was treated as such
            "id enim ait", "turbabuntur a facie eius patris or phanorum et iudicis uiduarum"
        ])
        result = tagger.tag_str(
            "( id enim ait) turbabuntur a facie eius patris or phanorum et iudicis uiduarum ..",
            processor=processor,
            iterator=data_iterator
        )
        tokens = [t["form"] for t in result]
        self.assertEqual(
            ["(", "id", "enim", "ait", ")", "turbabuntur", "a", "facie", "eius", "patris", "or", "phanorum",
             "et", "iudicis", "uiduarum", ".", "."],
            tokens,
            "Leading punctuation should not break anything"
        )

    def test_punctuation_is_not_seen(self):
        """Check that punctuation is not seen by the tagger

        """
        tagger, data_iterator, processor = make_controller([
            "id enim ait", "turbabuntur a facie eius patris or phanorum et iudicis uiduarum"
        ])
        tagger.tag_str(
            "( id enim ait ) turbabuntur a facie eius patris or phanorum et iudicis uiduarum .  .",
            processor=processor,
            iterator=data_iterator
        )
        self.assertNotIn(
            "(",
            [tok for sent in tagger.seen for tok in sent],
            "Punctuation should not be seen by the Tagger"
        )

    def test_j_are_temporarly_replaced(self):
        """Check that characters are replaced for the tagger, thus avoiding out of domain, and reinserted

        """
        tagger, data_iterator, processor = make_controller([
            "iudicis uiduarum"
        ])
        result = tagger.tag_str(
            "judicis uiduarum",
            processor=processor,
            iterator=data_iterator
        )
        flatten_seen = list([tok for sent in tagger.seen for tok in sent])

        self.assertEqual(result[0]["form"], "judicis", "'j' should be removed from tagging")
        self.assertEqual(result[0]["treated"], "iudicis", "And 'i' should replace it")

    def test_underscores(self):
        string = "una operatio in ecclesiae fundamento.._... _ . laetatur autem pater quia filius perierat"
        tagger, data_iterator, processor = make_controller([
            "una operatio in ecclesiae fundamento", "laetatur autem pater quia filius perierat"
        ])
        tagger.tag_str(
            string,
            processor=processor,
            iterator=data_iterator
        )
        flatten_seen = list([tok for sent in tagger.seen for tok in sent])
        self.assertEqual(
            ['una', 'operatio', 'in', 'ecclesiae', 'fundamento', 'laetatur', 'autem', 'pater', 'quia', 'filius',
             'perierat'],
            flatten_seen,
            "Seen element should not count the underscord"
        )

    def test_with_fake_advanced_tagger(self):
        target = write_crazy_file()
        tagger, it, pr = create_auto_tagger(lasla, lower=True)
        out_file = tagger.tag_file(target, it, pr)
        content = []
        with open(out_file) as f:
            header = []
            for line in f:
                splitted = line.strip().split()
                if not header:
                    header = splitted
                    continue
                content.append(dict(list(zip(header, splitted))))
        self.maxDiff = None

        fam = FakeAnnotationMaker()
        """\\\<1>[$@$](V)\\\§
\\<1>[$@$]\\§
§
\[I]\§
At o sceleste penis, o meum malum,
graui piaque lege noxiam lues.
Licet querare, nec tibi tener puer"""
        self.assertEqual(
                content,
            [
                *fam.make_multiple_response(["\\", "\\", "\\", "<"], is_ignored=True),
                fam.make_response("1"),
                *fam.make_multiple_response([">", "[", "$", "@", "$", "]", "("], is_ignored=True),
                fam.make_response("V", is_ignored=False, treated="3"),  # Roman Numeral
                *fam.make_multiple_response(
                    [")", "\\", "\\", "\\", "§", "\\", "\\", "<"], is_ignored=True),
                fam.make_response("1"),
                *fam.make_multiple_response([
                    ">", "[", "$", "@", "$", "]",
                    "\\", "\\", "§",
                    "§",
                    "\\", "["], is_ignored=True),
                fam.make_response("I", is_ignored=False, treated="1"),  # Roman Numeral
                *fam.make_multiple_response(["]", "\\", "§"], is_ignored=True),
                *fam.make_multiple_response(["At", "o", "sceleste", "penis"]),
                fam.make_response(",", is_ignored=True),
                *fam.make_multiple_response(["o", "meum", "malum"]),
                fam.make_response(",", is_ignored=True),
                *fam.make_multiple_response(["graui", "piaque", "lege", "noxiam", "lues"]),
                fam.make_response(".", is_ignored=True),
                *fam.make_multiple_response(["Licet", "querare"]),
                fam.make_response(",", is_ignored=True),
                *fam.make_multiple_response(["nec", "tibi", "tener", "puer", ])
            ]
        )

    def test_normalizers(self):
        "Check that references and abbreviations are correctly ignored"
        iterator, _ = get_iterator_and_processor()

        self.assertEqual(
            list(iterator("[REF:1.a.b]Quis est M. Cicero ? [REF:1.a.c] Ego sum !")),
            [
                (['Quis', 'est', 'M', 'Cicero'], 4, {0: '[REF:1.a.b]', 5: '?'}),
                (['Ego', 'sum'], 2, {0: '[REF:1.a.c]', 3: '!'})
            ],
            "References and abbreviations are kept as tokens. Abbreviations are sent in without DOT,"
            "References are untouched"
        )

        self.assertEqual(
            iterator.tokenizer.tokens,
            [(0, '[REF:1.a.b]', '[REF:1.a.b]'), (1, 'Quis', 'Quis'), (2, 'est', 'est'), (3, 'M.', 'M'),
             (4, 'Cicero', 'Cicero'), (5, '?', '?'), (6, '[REF:1.a.c]', '[REF:1.a.c]'), (7, 'Ego', 'Ego'),
             (8, 'sum', 'sum'), (9, '!', '!')],
            "Memory should be kept for abbreviations"
        )




from .processor import FroRulesProcessor, FroGlueProcessor
from pie_extended.pipeline.postprocessor.proto import RenamedTaskProcessor
from .tokenizer import FroMemorizingTokenizer
from pie_extended.pipeline.iterators.proto import DataIterator, GenericExcludePatterns
from pie_extended.pipeline.postprocessor.memory import MemoryzingProcessor


def get_iterator_and_processor(max_tokens=256):
    tokenizer = FroMemorizingTokenizer()
    processor = FroRulesProcessor(
        apply_on_reinsert=True,
        head_processor=MemoryzingProcessor(
            tokenizer_memory=tokenizer,
            head_processor=FroGlueProcessor(
                head_processor=RenamedTaskProcessor({"pos": "POS", "NOMB": "NOMB.", "PERS": "PERS."})
            )
        )
    )
    iterator = DataIterator(
        tokenizer=tokenizer,
        exclude_patterns=[GenericExcludePatterns.Punctuation_and_Underscore],
        max_tokens=max_tokens
    )
    return iterator, processor




from ...utils import Metadata, File, get_path

DESC = Metadata(
        "Middle Dutch",
        "dum",
        ["Mike Kestemont"],
        "Middle Dutch Model",
        "https://github.com/hipster-philology/middle-dutch-model"
)

VERSION = "0.0.1"
DOWNLOADS = [
    File(f"https://github.com/hipster-philology/middle-dutch-model/releases/download/{VERSION}/lemma.tar", "lemma.tar"),
    File(f"https://github.com/hipster-philology/middle-dutch-model/releases/download/{VERSION}/pos.tar", "pos.tar")
]


Models = "<{},lemma><{},pos>".format(
        get_path("dum", "lemma.tar"),
        get_path("dum", "pos.tar")
)



from ...utils import Metadata, File, get_path

DESC = Metadata(
        "Early Modern French",
        "freem",
        ["Simon Gabay", "Jean-Baptiste Camps", "Thibault Clérice", "Jean-Baptiste Tanguy", "Matthias Gille-Levenson",
         "Florian Cafiero"],
        "Standardizing linguistic data: method and tools for annotating(pre-orthographic) French",
        "https://github.com/e-ditiones/LEM17"
)

DOWNLOADS = [
        File("https://github.com/e-ditiones/LEM17/raw/7ec86d61c9d13cc8d0e56d6dd37971f0f220ad70/"
             "Models/Train_2/models/lemma.tar", "lemma.tar"),
        File("https://github.com/e-ditiones/LEM17/raw/7ec86d61c9d13cc8d0e56d6dd37971f0f220ad70/"
             "Models/Train_2/models/pos.tar", "pos.tar")
]


Models = "<{},lemma><{},POS>".format(
        get_path("freem", "lemma.tar"),
        get_path("freem", "pos.tar")
)



import os
from typing import Optional, Dict, Generator, Type

from pie.utils import shutup

with shutup():
    from pie.tagger import Tagger
    from pie import utils

from .pipeline.formatters.proto import Formatter
from .pipeline.disambiguators.proto import Disambiguator
from .pipeline.iterators.proto import DataIterator
from .pipeline.postprocessor.proto import ProcessorPrototype


class ExtensibleTagger(Tagger):
    def __init__(self, device='cpu', batch_size=100, lower=False, disambiguation=None):
        super(ExtensibleTagger, self).__init__(
            device=device,
            batch_size=batch_size,
            lower=True
        )
        self.disambiguation: Optional[Disambiguator] = disambiguation

    def tag_file(self, fpath: str, iterator: DataIterator, processor: ProcessorPrototype, no_tokenizer: bool = False):
        # Read content of the file
        with open(fpath) as f:
            data = f.read()

        _, ext = os.path.splitext(fpath)

        out_file = utils.ensure_ext(fpath, ext, 'pie')
        with open(out_file, 'w+') as f:
            for line in self.iter_tag(data, iterator, processor=processor, no_tokenizer=no_tokenizer):
                f.write(line)

        return out_file

    def tag_str(self, data: str, iterator: DataIterator, processor: ProcessorPrototype,
                no_tokenizer: bool = False) -> str:
        return list(self.iter_tag_token(data, iterator, processor=processor, no_tokenizer=no_tokenizer))

    def iter_tag_token(self, data: str, iterator: DataIterator, processor: ProcessorPrototype,
                       no_tokenizer: bool = False) -> Generator[Dict[str, str], None, None]:
        # Reset at each document
        processor.reset()
        iterator.tokenizer.reset()
        # Iterate !
        for chunk in utils.chunks(
                iterator(data, lower=self.lower, no_tokenizer=no_tokenizer),
                size=self.batch_size):

            # Unzip the batch into the sentences, their sizes and the dictionaries of things that needs
            #  to be reinserted
            sents, lengths, needs_reinsertion = zip(*chunk)
            is_empty = [not bool(sent) for sent in sents]

            tagged, tasks = self.tag(
                sents=[sent for sent in sents if sent],
                lengths=[l for l in lengths if l != 0]
            )

            if not processor.task_init:
                processor.set_tasks(tasks)

            # We keep a real sentence index
            for sents_index, sent_is_empty in enumerate(is_empty):
                if sent_is_empty:
                    sent = []
                else:
                    sent = tagged.pop(0)

                # Gets things that needs to be reinserted
                sent_reinsertion = needs_reinsertion[sents_index]

                # If we have a disambiguator, we run the results into it
                if self.disambiguation and sent:
                    sent = self.disambiguation(sent, tasks)

                reinsertion_index = 0

                for index, (token, tags) in enumerate(sent):
                    # Before current index
                    while reinsertion_index + index in sent_reinsertion:
                        yield processor.reinsert(sent_reinsertion[reinsertion_index+index])
                        del sent_reinsertion[reinsertion_index + index]
                        reinsertion_index += 1

                    yield from processor.get_dict(token, tags)

                for reinsertion in sorted(list(sent_reinsertion.keys())):
                    yield processor.reinsert(sent_reinsertion[reinsertion])

    def iter_tag(self, data: str, iterator: DataIterator, processor: ProcessorPrototype,
                 formatter_class: Type[Formatter] = Formatter, no_tokenizer: bool = False):
        formatter = None

        for annotation in self.iter_tag_token(data, iterator, processor, no_tokenizer = no_tokenizer):
            if not formatter:
                formatter = formatter_class(processor.tasks)
                yield formatter.write_headers()
            yield formatter.write_line(formatter.format_line(annotation))

        if formatter:
            yield formatter.write_footer()



from ...utils import Metadata, File, get_path

DESC = Metadata(
        "Ancient Greek",
        "grc",
        ["Thibault Clérice", "Enrique Manjavacas"],
        "Ancient Greek Model",
        "https://github.com/hipster-philology/greek-lemmatization-data"
)

DOWNLOADS = [
    File("https://github.com/hipster-philology/greek-lemmatization-data/raw/10eea6a2bf4aa1690844a28a8f1e40bc6a898b02/"
         "greek.lemma.tar", "lemma.tar"),
    File("https://github.com/hipster-philology/greek-lemmatization-data/raw/10eea6a2bf4aa1690844a28a8f1e40bc6a898b02/"
         "greek.pos.tar", "pos.tar"),
]


Models = "<{},lemma><{},pos>".format(
        get_path("grc", "lemma.tar"),
        get_path("grc", "pos.tar")
)



from typing import List, Dict, Optional
from .proto import ChainedProcessor, ProcessorPrototype
from copy import deepcopy


class SplitterPostProcessor(ChainedProcessor):
    """ This processor aims to help in splitting enclitics and proclitics


    >>> proc = SplitterPostProcessor()
    >>> proc.set_tasks(["lemma", "POS", "Case"])
    ['lemma', 'POS', 'Case']
    >>> proc.get_dict("praedicatione", ["praedicatio界-ne", "NOMcom", "Nominatif"]) == [
    ...    {"form": "praedicatione", "lemma": "praedicatio", "POS": "NOMcom", "Case": "Nominatif"},
    ...    {"form": "praedicatione", "lemma": "-ne", "POS": "NOMcom", "Case": "Nominatif"},
    ... ]
    True

    `keep=True` can be used with something like a RuleBased tokenizer to detect beforehand clitics
    and treat the tokens for example with RuleBased on top of it (len(get_dict()) > 1)

    >>> proc = SplitterPostProcessor(keep=True)
    >>> proc.set_tasks(["lemma", "POS", "Case"])
    ['lemma', 'POS', 'Case']
    >>> proc.get_dict("praedicatione", ["praedicatio界-ne", "NOMcom", "Nominatif"]) == [
    ...    {"form": "praedicatione", "lemma": "praedicatio界-ne", "POS": "NOMcom", "Case": "Nominatif"},
    ...    {"form": "praedicatione", "lemma": "praedicatio界-ne", "POS": "NOMcom", "Case": "Nominatif"}
    ... ]
    True
    >>> proc = SplitterPostProcessor(keep=False, prefix="界")
    >>> proc.set_tasks(["lemma", "POS", "Case"])
    ['lemma', 'POS', 'Case']
    >>> proc.get_dict("praedicatione", ["praedicatio界ne", "NOMcom", "Nominatif"]) ==  [
    ...    {"form": "praedicatione", "lemma": "praedicatio", "POS": "NOMcom", "Case": "Nominatif"},
    ...    {"form": "praedicatione", "lemma": "界ne", "POS": "NOMcom", "Case": "Nominatif"}
    ... ]
    True
    """

    def __init__(self, split_char: str = "界", column: str = "lemma", keep: bool = False,
                 head_processor: Optional[ProcessorPrototype] = None, prefix: str = "", **kwargs):
        super(SplitterPostProcessor, self).__init__(head_processor, **kwargs)
        self.split_char: str = split_char
        self.column: str = column
        self.keep: bool = keep
        self.prefix: str = prefix

    def get_dict(self, token: str, tags: List[str]) -> List[Dict[str, str]]:
        out = []
        for anno in super(SplitterPostProcessor, self).get_dict(token=token, tags=tags):
            if self.split_char in anno[self.column]:
                for number, new_val in enumerate(anno[self.column].split(self.split_char)):
                    if number > 0:
                        new_val = self.prefix + new_val
                    if self.keep:
                        out.append(anno)
                    else:
                        new = deepcopy(anno)
                        new[self.column] = new_val
                        out.append(new)
            else:
                out.append(anno)
        return out



from typing import Tuple, Match
import regex as re

from pie_extended.pipeline.tokenizers.utils.excluder import ExcluderPrototype, DASH, APOSTROPHE
from pie_extended.pipeline.tokenizers.utils import chars


AUJOURDHUI_MASK = "字"
CLITICS = tuple(sorted([
    '-ce', '-ci', '-elle', '-elles', '-en', '-eux', '-il', '-ils', '-je', '-la', '-le',
    '-les', '-leur', '-leurs',
    '-lui', '-là', '-m', '-me', '-moi', '-même', '-mêmes', '-nous', '-on', '-t',
    '-te', '-toi', '-tu', '-un',
    '-une', '-unes', '-uns', '-vous', '-y',
    # No idea what was the goal of the following line...
    # r'-m字\d', r'-t字\d'  # 字 replaces all kinds of apostrophe
], key=len, reverse=True))

# Abbreviations separated by `; ` (SEMICOLON+SPACE)
# You can use regexp inside this list, such as [Cc]f. but I would rather recommend to add both `cf.; Cf;`
#   for readability.
ABBREVIATIONS = sorted(list(set(
    "Acad.; Adj.; Agricol.; Agricul.; Apocal.; anc.; Bot.; Botan.; Botaniq.; ca.; cap.; capi.; cf.; Cf.; " 
    "Cha.; Chap.; Col.; Dic.; Diction.; Dictionn.; Eccl.; Écon.; Élem.; Fig.; Fr.; Geog.; Gram.; " 
    "Gramm.; Hist.; Ibid.; Ibid.; Inst.; Jard.; Jurisprud.; Latit.; Li.; Lib.; Libr.; Lig.; Lit.; " 
    "Littérat.; Liv.; Long.; Mar.; Mat.; Mathém.; Mech.; Med.; Med.; Mem.; Menuis.; Milit.; Mod.; " 
    "Mor.; Mr.; Monsr.; nat.; natur.; N.b.; Orat.; Ornith.; Ornythol.; Ornitholog.; Part.; Pag.; " 
    "Pharm.; Phil.; Philos.; Pl.; Pl.; Politiq.; P.S.; Phys.; Physiq.; Sr.; St.; Subst.; s.f.; " 
    "S.M.; s.m.; Tab.; Théât.; Trév.; Tom.; Vol.; V. n.; V. a.; V. act.; Zoo.; " 
    "Zoolog.".split("; ")
)), key=len, reverse=True)


class AujourdhuiExcluder(ExcluderPrototype):
    """ Normalizes Aujourd'hui before it goes into splitting sentences

    >>> excl = AujourdhuiExcluder()
    >>> excl.before_sentence_tokenizer("Aujourd'hui j'ai poney")
    "Aujourd字0hui j'ai poney"
    >>> excl.after_sentence_tokenizer("Aujourd字0hui j'ai poney")
    "Aujourd'hui j'ai poney"
    """
    def __init__(self, apostrophe: str = chars.APOSTROPHE, apostrophe_mask: str = AUJOURDHUI_MASK):
        self.apostrophes = apostrophe
        self.mask = apostrophe_mask
        self.re = re.compile("(aujourd)(["+self.apostrophes+"])(hui)", flags=re.IGNORECASE)

    def _replace(self, regex_match: Match) -> str:
        return regex_match.group(1) + self.mask + str(self.apostrophes.index(regex_match.group(2))) + regex_match.group(3)

    def before_sentence_tokenizer(self, value: str) -> str:
        return self.re.sub(self._replace, value)

    def after_sentence_tokenizer(self, value: str) -> str:

        for index_apo, apostrophe in enumerate(self.apostrophes):
            value = value.replace(self.mask+str(index_apo), apostrophe)
        return value


class FrenchCliticsExcluder(ExcluderPrototype):
    def __init__(self, clitics: Tuple[str, ...] = CLITICS, dash: str = DASH,
                 add_space_before: bool = True, add_space_after: bool = True,
                 apostrophes: str = chars.APOSTROPHE, apostrophes_mask: str = APOSTROPHE):
        self.clitics = clitics
        self.dash = dash
        self.apostrophes = apostrophes
        self.apostrophes_mask = apostrophes_mask
        self.re = re.compile(
            r"((-)(t))?(-)(" +
            r"|".join([tok.replace("-", "") for tok in self.clitics]) +
            # We use apostrophe_mask and the \d because a mask could have already been applied
            r")([" + self.apostrophes + r"]|" + self.apostrophes_mask + r"\d|\b|$)",
            flags=re.IGNORECASE
        )

        # Space handling
        self.space_before: str = ""
        if add_space_before:
            self.space_before = " "

        self.space_after: str = ""
        if add_space_after:
            self.space_after = " "

    def _replace_clitic(self, match: Match) -> str:
        # Group 1: - (of -t)
        # Group 2: t (of -t)
        # Group 3: - (of clitic)
        # Group 4: t (the clitic)
        # Group 5: apostrophe or word boundary
        g0, g1, g2, g3, g4, g5 = match.groups()
        before = ""
        if g0:
            before = self.dash+"t"

        return self.space_before + before + self.dash + g4 + g5 + self.space_after

    def before_sentence_tokenizer(self, value: str) -> str:
        """ Normalize before sentence is tokenized

        >>> excl = FrenchCliticsExcluder()
        >>> excl.before_sentence_tokenizer("A-t-il mangé le peut-être que moi-même je connais ?")
        'A 精t精il  mangé le peut-être que moi 精même  je connais ?'

        Note that it introduces a secondary space.

        >>> excl.before_sentence_tokenizer("Va-t'en va-nu-pieds !")
        "Va 精t' en va-nu-pieds !"

        """
        return self.re.sub(
            self._replace_clitic,
            value
        )

    def after_sentence_tokenizer(self, value: str) -> str:
        return value.replace(self.dash, "-")



import regex as re
from typing import Dict, Pattern

from pie_extended.pipeline.postprocessor.glue import GlueProcessor
from pie_extended.pipeline.postprocessor.rulebased import RuleBasedProcessor
from pie_extended.pipeline.postprocessor.proto import RenamedTaskProcessor


class FroRulesProcessor(RuleBasedProcessor):
    """ Fro Dataset has not all punctuation signs in it, we remove it and posttag it automatically

    """
    PONCTU: Pattern = re.compile(r"^\W+$")
    NUMBER: Pattern = re.compile(r"\d+")
    PONFORT = [".", "...", "!", "?"]

    def rules(self, annotation: Dict[str, str]) -> Dict[str, str]:
        token = annotation["form"]
        if self.PONCTU.match(token):
            if token in self.PONFORT:
                pos = "PONfrt"
            else:
                pos = "PONfbl"
            return {"form": token, "lemma": token, "POS": pos, "morph": "MORPH=empty", "treated": token}
        elif self.NUMBER.match(token):
            annotation["pos"] = "ADJcar"
        return annotation

    def __init__(self, *args, **kwargs):
        super(FroRulesProcessor, self).__init__(*args, **kwargs)


class FroGlueProcessor(GlueProcessor):
    """ We glue morphological features into one column

    """
    OUTPUT_KEYS = ["form", "lemma", "POS", "morph"]
    GLUE = {"morph": ["MODE", "TEMPS", "PERS.", "NOMB.", "GENRE", "CAS", "DEGRE"]}
    EMPTY_TAG: Dict[str, str] = {"CAS": "CAS=x",
                                 "NOMB.": "NOMB.=x",
                                 "DEGRE": "DEGRE=x",
                                 "MODE": "MODE=x",
                                 "TEMPS": "TEMPS=x",
                                 "GENRE": "GENRE=x",
                                 "PERS.": "PERS.=x"}

    def __init__(self, *args, **kwargs):
        super(FroGlueProcessor, self).__init__(*args, **kwargs)
    
    def _get_glued(self, glued_task: str, token_dict: Dict[str, str]):
        return token_dict[glued_task]  # Tempfix because tasks contains their own name



from .tokenizer import GrcMemorizingTokenizer
from pie_extended.pipeline.iterators.proto import DataIterator
from pie_extended.pipeline.postprocessor.proto import ProcessorPrototype
from pie_extended.pipeline.postprocessor.memory import MemoryzingProcessor


def get_iterator_and_processor(max_tokens=256):
    tokenizer = GrcMemorizingTokenizer()
    processor = MemoryzingProcessor(
        tokenizer_memory=tokenizer,
        head_processor=ProcessorPrototype()
    )

    iterator = DataIterator(
        tokenizer=tokenizer,
        max_tokens=max_tokens
    )
    return iterator, processor




from ...utils import Metadata, File, get_path

DESC = Metadata(
        "Français classique et moderne",
        "fr",
        ["Jean-Baptiste Camps", "Simon Gabay", "Thibault Clérice", "Florian Cafiero"],
        "Corpus and Models for Lemmatisation and POS-tagging of Classical French Theatre",
        "TODO: add arXiv link"
)

DOWNLOADS = [
        File("https://zenodo.org/record/3701294/files/fr-class-wembs-lemma-2020_03_06-01_52_24.tar?download=1", "lemma.tar"),
        File("https://zenodo.org/record/3701320/files/fr-class-pos-wembs_aux-pos-2020_03_03-18_52_52.tar?download=1", "pos.tar"),
        File("https://zenodo.org/record/3701320/files/fr-class-morph-wembs_aux-MODE-2020_03_08-17_46_05.tar?download=1", "mode.tar"),
        File("https://zenodo.org/record/3701320/files/fr-class-morph-wembs_aux-TEMPS-2020_03_08-19_03_26.tar?download=1", "temps.tar"),
        File("https://zenodo.org/record/3701320/files/fr-class-morph-wembs_aux-PERS-2020_03_08-18_24_29.tar?download=1", "pers.tar"),
        File("https://zenodo.org/record/3701320/files/fr-class-morph-wembs_aux-NOMB-2020_03_08-17_55_50.tar?download=1", "nomb.tar"),
        File("https://zenodo.org/record/3701320/files/fr-class-morph-wembs_aux-GENRE-2020_03_08-16_54_44.tar?download=1", "genre.tar"),
        File("https://zenodo.org/record/3701320/files/fr-class-morph-wembs_aux-CAS-2020_03_08-16_23_57.tar?download=1", "cas.tar")
]


Models = "<{},lemma><{},pos><{},MODE><{},TEMPS><{},PERS><{},NOMB><{},GENRE><{},CAS>".format(
        get_path("fr", "lemma.tar"),
        get_path("fr", "pos.tar"),
        get_path("fr", "mode.tar"),
        get_path("fr", "temps.tar"),
        get_path("fr", "pers.tar"),
        get_path("fr", "nomb.tar"),
        get_path("fr", "genre.tar"),
        get_path("fr", "cas.tar")
)



from pie_extended.pipeline.postprocessor.proto import ProcessorPrototype
from pie_extended.pipeline.iterators.proto import DataIterator
from pie_extended.pipeline.tokenizers.simple_tokenizer import LengthTokenizer


def get_iterator_and_processor(max_tokens=256):
    tokenizer = LengthTokenizer(35)
    processor = ProcessorPrototype()
    iterator = DataIterator(
        tokenizer=tokenizer,
        max_tokens=max_tokens
    )
    return iterator, processor




from typing import List, Dict, Optional, Type, Generator, Tuple

DEFAULT_EMPTY = "_"


class ProcessorPrototype:
    empty_value: str

    def __init__(self, empty_value: Optional[str] = None):
        """ Applies postprocessing. Simplest Processor one could use.

        :param empty_value: Value to use to fill tasks that would not get any data


        >>> x = ProcessorPrototype(empty_value="%")
        >>> x.set_tasks(["a", "b"])
        ['a', 'b']
        >>> x.reinsert("x") == {"form": "x", "a": "%", "b": "%"}
        True
        >>> x.get_dict("y", ["1", "2"]) == [{"form": "y", "a": "1", "b": "2"}]
        True
        """
        self._tasks = []
        self.task_init: bool = False
        self.empty_value = empty_value or DEFAULT_EMPTY

    @property
    def tasks(self) -> List[str]:
        return self._tasks

    def set_tasks(self, tasks) -> List[str]:
        self.task_init = True
        self._tasks = tasks
        return tasks

    def postprocess(self, line):
        pass

    def reinsert(self, form: str) -> Dict[str, str]:
        """ Generates an automatic line for a token that was removed from lemmatization

        :param form: Token to reinsert
        :return: Dictionary representation of the token, as an annotation


        >>> x = ProcessorPrototype(empty_value="%")
        >>> x.set_tasks(["a", "b"])
        ['a', 'b']
        >>> x.reinsert("x") == {"form": "x", "a": "%", "b": "%"}
        True
        """
        return dict(form=form, **{task: self.empty_value for task in self._tasks})

    def get_dict(self, token: str, tags: List[str]) -> List[Dict[str, str]]:
        """ Get the dictionary representation of a token annotation

        :param token: Token used as input for pie
        :param tags: List of tags generated
        :return: Dictionary representation of the token and its annotations

        >>> x = ProcessorPrototype(empty_value="%")
        >>> x.set_tasks(["a", "b"])
        ['a', 'b']
        >>> x.get_dict("y", ["1", "2"]) == [{"form": "y", "a": "1", "b": "2"}]
        True
        """
        return [{"form": token, **{k: val for k, val in zip(self._tasks, tags)}}]

    def reset(self):
        """ Functions that should be run in between documents

        >>> x = ProcessorPrototype(empty_value="%")
        >>> x.set_tasks(["a", "b"])
        ['a', 'b']
        >>> x.reset()
        """
        pass


class RenamedTaskProcessor(ProcessorPrototype):
    def __init__(self, task_map: Dict[str, str], **kwargs):
        """ This Processor is used for renaming tasks (Pie for example refuses tasks containing dots)

        >>> x = RenamedTaskProcessor({"task_name_1": "renamed"})
        >>> x.set_tasks(["task_name_1", "y"])
        ['renamed', 'y']
        >>> x.get_dict("token", ["a", "b"]) == [{"form": "token", "renamed": "a", "y": "b"}]
        True
        """
        super(RenamedTaskProcessor, self).__init__(**kwargs)
        self._map: Dict[str, str] = task_map

    def set_tasks(self, tasks):
        self._tasks = [self._map.get(task, task) for task in tasks]
        return self.tasks


class ChainedProcessor(ProcessorPrototype):
    """ Allows for easy chaining !

    The ChainedProcessor is basically using its headprocessor in the background and checking it's output to some extent

    The prototype of ChainedProcessor using Processor Prototype would have the same results because
    chained processor is not doing anything new except enabling chaining

        >>> x = ProcessorPrototype(empty_value="%")
        >>> x.set_tasks(["a", "b"])
        ['a', 'b']
        >>> y = ChainedProcessor(x)
        >>> y.set_tasks(["a", "b"])
        ['a', 'b']
        >>> x.reinsert("x") == y.reinsert("x")
        True
        >>> x.get_dict("y", ["1", "2"]) == y.get_dict("y", ["1", "2"])
        True

    You can subclass it to modify the output of the preceding processor :

    >>> class ExampleChained(ChainedProcessor):
    ...     def reinsert(self, form: str) -> Dict[str, str]:
    ...         annotation = self.head_processor.reinsert(form)
    ...         annotation["col3"] = "x"
    ...         return annotation
    ...
    ...     def get_dict(self, form: str, tags: List[str]) -> List[Dict[str, str]]:
    ...         return [{"col3": "x", **{x:y for x, y in anno.items() if x != "col3"}}
    ...                  for anno in self.head_processor.get_dict(form, tags)]
    >>> x = ExampleChained(ProcessorPrototype(empty_value="EMPTY"))
    >>> x.set_tasks(["a", "b"])
    ['a', 'b']
    >>> x.reinsert("x") == {"form": "x", "a": "EMPTY", "b": "EMPTY", "col3": "x"}
    True
    >>> x.get_dict("y", ["1", "2"]) == [{"form": "y", "a": "1", "b": "2", "col3": "x"}]
    True

    """
    head_processor: ProcessorPrototype

    def __init__(self, head_processor: Optional[ProcessorPrototype], **kwargs):
        super(ChainedProcessor, self).__init__(**kwargs)

        self.head_processor: ProcessorPrototype = head_processor
        if not self.head_processor:
            self.head_processor = ProcessorPrototype()

    def set_tasks(self, tasks):
        self._tasks = self.head_processor.set_tasks(tasks)
        return self.tasks

    def reinsert(self, form: str) -> Dict[str, str]:
        return self.head_processor.reinsert(form)

    def get_dict(self, token: str, tags: List[str]) -> List[Dict[str, str]]:
        return self.head_processor.get_dict(token, tags)

    def reset(self):
        self.head_processor.reset()



import regex as re
from typing import List, Generator, Tuple
from unidecode import unidecode

from pie_extended.models.fro.tokenizer import _Dots_except_apostrophe
from pie_extended.pipeline.tokenizers.memorizing import MemorizingTokenizer
from pie_extended.models.lasla._params import abbrs
from pie_extended.pipeline.tokenizers.utils.excluder import (
    ReferenceExcluder,
    ExcluderPrototype,
    RegexpExcluder,
    AbbreviationsRemoverExcluder,
    DEFAULT_CHAR_REGISTRY
)
from pie_extended.pipeline.tokenizers.utils import regexps
from pie_extended.utils import roman_number


class LatMemorizingTokenizer(MemorizingTokenizer):
    re_add_space_around_punct = re.compile(r"(\s*)([^\w\s])(\s*)")
    re_words = re.compile("^\w+$")
    _sentence_boundaries = re.compile(
        r"([" + _Dots_except_apostrophe + r"]+\s*)+"
    )
    re_roman_number = re.compile(r"^"+regexps.RomanNumbers+"$")

    def __init__(self):
        super(LatMemorizingTokenizer, self).__init__()
        self.tokens = []
        self.char_registry = DEFAULT_CHAR_REGISTRY
        self.normalizers: Tuple[ExcluderPrototype, ...] = (
            ReferenceExcluder(char_registry=self.char_registry),
            ReferenceExcluder(regex_string=r"(\[IGN:[^\]]+\])", char_registry=self.char_registry),
            RegexpExcluder(r"(\p{Greek}+)"),
            RegexpExcluder(r"(\p{No})"),
            AbbreviationsRemoverExcluder(abbrs=abbrs, char_registry=self.char_registry)
        )

    @staticmethod
    def _sentence_tokenizer_merge_matches(match):
        """ Best way we found to deal with repeating groups"""
        start, end = match.span()
        return match.string[start:end] + "<SPLIT>"

    def _real_sentence_tokenizer(self, string: str) -> List[str]:
        string = self._sentence_boundaries.sub(self._sentence_tokenizer_merge_matches, string)

        for normalizer in self.normalizers:
            string = normalizer.after_sentence_tokenizer(string)

        return string.split("<SPLIT>")

    def _real_word_tokenizer(self, text: str, lower: bool = False) -> List[str]:
        if lower is True:
            text = text.lower()
        return text.split()

    def sentence_tokenizer(self, text: str, lower: bool = False) -> Generator[List[str], None, None]:
        """

        >>> x = LatMemorizingTokenizer()
        >>> list(x.sentence_tokenizer("XX Lasciva puella et lasciue C. Agamemnone whateverve."))
        [['3', 'Lasciua', 'puella', 'et', 'lasciue', 'C', 'Agamemnone', 'whateuerue', '.']]

        """
        sentences = list()
        data = self.normalizer(text)
        for sent in self._real_sentence_tokenizer(data):
            sent = sent.strip()
            if sent:
                sentences.append(self.word_tokenizer(sent))
        yield from sentences

    def normalizer(self, data: str) -> str:
        for excluder in self.normalizers:
            data = excluder.before_sentence_tokenizer(data)

        data = self.re_add_space_around_punct.sub(
            r" \g<2> ",
            data
        )
        return data

    def roman_to_number(self, inp: str) -> str:
        out = roman_number(inp)
        if out > 3:
            out = 3
        return str(out)

    def replacer(self, inp: str):
        """

        :param inp:
        :return:

        """
        for excluder in self.normalizers:
            if not excluder.can_be_replaced and excluder.exclude_regexp.match(inp):
                return inp
        if self.re_roman_number.match(inp):
            return self.roman_to_number(inp)
        elif inp.isnumeric():
            if inp.isdigit():  # avoid. ↀ
                if int(inp) > 3:
                    return "3"
                else:
                    return inp
            return "3"
        elif "." == inp:
            return "."

        if self.re_words.match(inp):  # Might check somehow else but for now... It's weird unidecode normalize into SS
            inp = unidecode(inp)
        inp = inp.replace("V", "U").replace("v", "u").replace("J", "I").replace("j", "i").replace(".", "")#.lower()
        return inp
